// Matrix exponential by pade approximation

// Based on C++ code by Dung-Bang Tsai
//  (which was released under the Boost Software License, Version 1.0
//                      and the GNU General Public License, Version 2)


class MatrixExp
{
public:
  MatrixExp(MyMatrix iH, const uint ip = 6)
    :n(iH.n_rows)
    ,p(ip)
    ,H(iH)
    ,I(iH.n_rows,iH.n_rows,fill::eye)
    ,hnorm(0.0)
    ,c(ip+1)
  {

#ifndef MOU_NO_DEBUG
    assert(H.is_square());
    assert(p >= 3);
#endif

    c[0]=1;
    for(uint i = 0; i < p; ++i)
      c[i+1] = c[i] * ((p - i)/((i + 1.0) * (2.0 * p - i)));
    for(uint i = 0; i < n; ++i)
      {
        double temp = 0.0;
	for(uint j = 0; j < n; j++)
	  temp += std::abs(H(i, j)); 
	hnorm = std::max<double>(hnorm, temp);
      }
  }

  MyMatrix operator()(double scale) const
  {
    double thnorm = std::abs(scale * hnorm);
    if (thnorm == 0.0)
      {
#ifndef MOU_NO_DEBUG
        assert(H.is_finite());
#endif
	return I;
      }
    
    int s = 0;
    if (thnorm >= 0.5)
      {
    	s = static_cast<int>((log(thnorm) / log(2.0))) + 2;
    	scale /= double(std::pow(2.0, s));
      }

    MyMatrix U = scale * H;
    MyMatrix H2 = U * U;
    MyMatrix Q = c[p] * H2 + c[p-2] * I;
    MyMatrix P = c[p-1] * H2 + c[p-3] * I;
    bool odd = true;
    for( uint k = p - 3; k > 0; --k)
      {
    	if (odd)
    	  Q = Q * H2 + c[k-1] * I;
    	else
    	  P = P * H2 + c[k-1] * I;
    	odd = !odd;
      }
    if (odd)
      {
        Q = Q * U;
	Q -= P;
	U = -2.0 * solve(Q, P) - I;
      }
    else
      {
        P = P * U;
        Q -= P;
	U =  2.0 * solve(Q, P) + I;
      }
    // Squaring 
    for(int i = 0; i < s; ++i)
      U = U * U;
    return U;
  }

private:
  uint n;
  uint p;
  MyMatrix H;
  MyMatrix I;
  double hnorm;
  vector<double> c;
};
