#include <cstdlib>
#include <cassert>
#include <cmath>
#include <vector>
#include <algorithm>
#include <limits>

#define ARMA_NO_DEBUG
#define MONOCAR_NO_DEBUG

#define USE_LONG

#define ANALYTIC_GRADIENT
#define ANALYTIC_HESSIAN

#define START_NULL

#ifdef USE_NLOPT
#include <nlopt.h>
#endif

#include "monocar.h"

using namespace arma;
using namespace std;

typedef arma::mat MyMatrix;
typedef arma::vec MyMatrixCol;
typedef arma::rowvec MyMatrixRow;
typedef arma::cube MyCube;
typedef arma::uword uint;

typedef Rcpp::NumericMatrix inputMatrix;
typedef Rcpp::NumericVector inputVector;
typedef Rcpp::IntegerVector inputIntegerVector;

#include "expm.h"


#define LOGTWOPI 1.83787706640934548356

int debugLevel;

int screenWidth;
int screenUse;


int maxIndex(const inputIntegerVector& input)
{
  int maximum = 0;
  for (inputIntegerVector::const_iterator ix = input.begin(); ix != input.end(); ix++)
    maximum = max(maximum, *ix);
  return maximum;
}

void printDot(const string& input)
{
  screenUse += input.size();
  if ((screenWidth>0) && (screenUse > screenWidth))
    {
      screenUse = input.size();
      Rcpp::Rcout << endl << input << flush;
    }
  else
    {
      Rcpp::Rcout << input << flush;
    }
}

MyMatrix vectorToMatrix(const vector<double>& input)
{
  MyMatrix output = conv_to<MyMatrix>::from(input);
  return(output);
}

MyMatrix RtoMatrix(inputMatrix& input)
{
  // MyMatrix output(input.begin(), input.nrow(), input.ncol());
  MyMatrix output=Rcpp::as<MyMatrix>(input);
  return(output);
}

MyMatrix RtoMatrix(inputVector& input)
{
  MyMatrix output(input.begin(), input.size(), 1);
  return(output);
}

MyMatrixCol RtoMatrixCol(inputVector& input)
{
  // MyMatrixCol output(input.begin(), input.size());
  MyMatrixCol output=Rcpp::as<MyMatrixCol>(input);
  return(output);
}

const MyMatrixCol RtoMatrixCol(const inputVector& input)
{
  //MyMatrixCol output(input.begin(), input.size(), false);
  MyMatrixCol output=Rcpp::as<MyMatrixCol>(input);
  return(output);
}

const MyMatrix RtoMatrix(const inputMatrix& input)
{
  //MyMatrix output(input.begin(), input.nrow(), input.ncol(), false);
  MyMatrix output=Rcpp::as<MyMatrix>(input);
  return(output);
}

const MyMatrix RtoMatrix(const inputVector& input)
{
  //MyMatrix output(input.begin(), input.size(), 1, false);
  MyMatrix output=Rcpp::as<MyMatrix>(input);
  return(output);
}

bool hasNegEigen(const MyMatrix& input)
{
  arma::cx_vec eigval = arma::eig_gen(input);
  for (arma::cx_vec::const_iterator ix = eigval.begin();
       ix != eigval.end();
       ix++)
    {
      if (real(*ix)<0)
	return true;
    }
  return false;
}


class MatrixExpFrechet
{
public:
  MatrixExpFrechet(const MyMatrix& iMatrix)
    :baseMatrix(iMatrix)
    ,n(iMatrix.n_rows)
  {
  }

  void push_back(const MyMatrix& iDMatrix)
  {
    bool iDnonzero = false;

    for(MyMatrix::const_iterator i=iDMatrix.begin(); i!=iDMatrix.end(); ++i)
      {
	if (*i != 0.0)
	  {
	    iDnonzero = true;
	    break;
	  }
      }

    if (iDnonzero)
      {
	MyMatrix xmat(2*n,2*n,fill::zeros);
	xmat.submat(0,0,size(n,n)) = baseMatrix;
	xmat.submat(n,n,size(n,n)) = baseMatrix;
	xmat.submat(0,n,size(n,n)) = iDMatrix;
	MatrixExp expmm(xmat);
	expm.push_back(expmm);
      }
    else
      {
	expm.push_back(baseMatrix);
      }

    nonzero.push_back(iDnonzero);
  }

  MyMatrix operator()(double scale, uint i) const
  {
    if (nonzero[i])
      return expm[i](scale).submat(0,n,size(n,n));
    else
      return MyMatrix(n,n,fill::zeros);
  }

  void operator()(MyMatrix& iBaseMatrix, MyMatrix& iDMatrix, double scale, uint i) const
  {
    if (nonzero[i])
      {
	iBaseMatrix = expm[i](scale).submat(0,0,size(n,n));
	iDMatrix = expm[i](scale).submat(0,n,size(n,n));
      }
    else
      {
	iBaseMatrix = expm[i](scale).submat(0,0,size(n,n));
	iDMatrix = MyMatrix(n,n,fill::zeros);
      }
  }

private:
  vector<bool> nonzero;
  vector<MatrixExp> expm;
  MyMatrix baseMatrix;
  uint n;
};


class MatrixExpDoubleFrechet
{
public:
  MatrixExpDoubleFrechet(const MyMatrix& iMatrix)
    :baseMatrix(iMatrix)
    ,n(iMatrix.n_rows)
  {
  }

  void push_back(const MyMatrix& iD1Matrix,
		 const MyMatrix& iD2Matrix,
		 const MyMatrix& iDDMatrix)
  {
    bool iDnonzero = false;

    for(MyMatrix::const_iterator i=iD1Matrix.begin(); i!=iD1Matrix.end(); ++i)
      {
	if (*i != 0.0)
	  {
	    iDnonzero = true;
	    break;
	  }
      }

    if (iDnonzero)
      {
	iDnonzero = false;
	for(MyMatrix::const_iterator i=iD2Matrix.begin(); i!=iD2Matrix.end(); ++i)
	  {
	    if (*i != 0.0)
	      {
		iDnonzero = true;
		break;
	      }
	  }
      }

    if (!iDnonzero)
      {
	for(MyMatrix::const_iterator i=iDDMatrix.begin(); i!=iDDMatrix.end(); ++i)
	  {
	    if (*i != 0.0)
	      {
		iDnonzero = true;
		break;
	      }
	  }
      }
    
    if (iDnonzero)
      {
	MyMatrix xmat(4*n,4*n,fill::zeros);
	xmat.submat(0,0,size(n,n)) = baseMatrix;
	xmat.submat(n,n,size(n,n)) = baseMatrix;
	xmat.submat(2*n,2*n,size(n,n)) = baseMatrix;
	xmat.submat(3*n,3*n,size(n,n)) = baseMatrix;
	xmat.submat(0,n,size(n,n)) = iD1Matrix;
	xmat.submat(2*n,3*n,size(n,n)) = iD1Matrix;
	xmat.submat(0,2*n,size(n,n)) = iD2Matrix;
	xmat.submat(n,3*n,size(n,n)) = iD2Matrix;
	xmat.submat(0,3*n,size(n,n)) = iDDMatrix;
	MatrixExp expmm(xmat);
	expm.push_back(expmm);
      }
    else
      {
	expm.push_back(baseMatrix);
      }

    nonzero.push_back(iDnonzero);
  }

  MyMatrix operator()(double scale, uint i) const
  {
    if (nonzero[i])
      return expm[i](scale).submat(0,3*n,size(n,n));
    else
      return MyMatrix(n,n,fill::zeros);
  }

  void operator()(MyMatrix& iBaseMatrix, MyMatrix& iD1D2Matrix, double scale, uint i) const
  {
    if (nonzero[i])
      {
	iBaseMatrix = expm[i](scale).submat(0,0,size(n,n));
	iD1D2Matrix = expm[i](scale).submat(0,3*n,size(n,n));
      }
    else
      {
	iBaseMatrix = expm[i](scale).submat(0,0,size(n,n));
	iD1D2Matrix = MyMatrix(n,n,fill::zeros);
      }
  }

private:
  vector<bool> nonzero;
  vector<MatrixExp> expm;
  MyMatrix baseMatrix;
  uint n;
};


MyMatrixCol rmvnorm(const MyMatrixCol mean, const MyMatrix vcov)
{
  MyMatrixCol epsilon(mean.n_rows, arma::fill::randn);
  if (arma::any(arma::vectorise(arma::abs(vcov))>1e-10))
    return mean + arma::chol(vcov) * epsilon;
  else
    return mean;
}


class OUparam
{
public:
  bool hasStart;
  MyMatrix start;
  MyMatrix mean;
  MyMatrix extraMean;
  MyMatrix reversion;
  MyMatrix sigma;
  MyMatrix extraVarAdd;
  MyMatrix extraVarMult;
  MyMatrix extraVarPow;
  MyMatrix exo;
  MyMatrixCol varCenters;
  bool hasExo;
  bool varByHouse;
  double getExtraVarAdd(const int houseIndex, const int seriesIndex) const
  {
    if (varByHouse)
      return extraVarAdd(houseIndex, 0);
    else
      return extraVarAdd(seriesIndex, 0);
  }
  double getExtraVarMult(const int houseIndex, const int seriesIndex) const
  {
    if (varByHouse)
      return extraVarMult(houseIndex, 0);
    else
      return extraVarMult(seriesIndex, 0);
  }
  double getExtraVarPow(const int houseIndex, const int seriesIndex) const
  {
    if (varByHouse)
      return extraVarPow(houseIndex, 0);
    else
      return extraVarPow(seriesIndex, 0);
  }
  double getVarCenter(const int houseIndex, const int seriesIndex) const
  {
    if (varByHouse)
      return varCenters(houseIndex);
    else
      return varCenters(seriesIndex);
  }
};


class OUparamspace{
public:
  OUparamspace(const int iProcessIndexSize,
	       const int iHouseSize,
	       const bool iHasStart,
	       const bool iHasExo,
	       const bool iVarByHouse,
	       const inputVector& iVarCenters,
	       const inputMatrix& iStartConstraint,
	       const inputMatrix& iMeanConstraint,
	       const inputMatrix& iExtraMeanConstraint,
	       const inputMatrix& iReversionConstraint,
	       const inputMatrix& iSimultaneousConstraint,
	       const inputMatrix& iSigmaSqrtConstraint,
	       const inputMatrix& iExtraVarAddConstraint,
	       const inputMatrix& iExtraVarMultConstraint,
	       const inputMatrix& iExtraVarPowConstraint,
	       const inputMatrix& iExoConstraint)
  :processIndexSize(iProcessIndexSize)
  ,houseSize(iHouseSize)
  ,hasStart(iHasStart)
  ,hasExo(iHasExo)
  ,varByHouse(iVarByHouse)
  ,varCenters(RtoMatrixCol(iVarCenters))
  ,startConstraint(RtoMatrix(iStartConstraint))
  ,meanConstraint(RtoMatrix(iMeanConstraint))
  ,extraMeanConstraint(RtoMatrix(iExtraMeanConstraint))
  ,reversionConstraint(RtoMatrix(iReversionConstraint))
  ,simultaneousConstraint(RtoMatrix(iSimultaneousConstraint))
  ,sigmaSqrtConstraint(RtoMatrix(iSigmaSqrtConstraint))
  ,extraVarAddConstraint(RtoMatrix(iExtraVarAddConstraint))
  ,extraVarMultConstraint(RtoMatrix(iExtraVarMultConstraint))
  ,extraVarPowConstraint(RtoMatrix(iExtraVarPowConstraint))
  ,exoConstraint(RtoMatrix(iExoConstraint))
  ,paramSize(iMeanConstraint.ncol())
  {
#ifndef MONOCAR_NO_DEBUG
    assert(iReversionConstraint.ncol() == paramSize);
    assert(iSimultaneousConstraint.ncol() == paramSize);
    assert(iSigmaSqrtConstraint.ncol() == paramSize);
    assert(iExtraVarAddConstraint.ncol() == paramSize);
    assert(iExtraVarMultConstraint.ncol() == paramSize);
    assert(iExtraVarPowConstraint.ncol() == paramSize);
    assert(uint(iReversionConstraint.nrow()) ==
	   uint(std::sqrt(iReversionConstraint.nrow()))*
	   uint(std::sqrt(iReversionConstraint.nrow())));
    assert(uint(iSimultaneousConstraint.nrow()) ==
	   uint(std::sqrt(iSimultaneousConstraint.nrow()))*
	   uint(std::sqrt(iSimultaneousConstraint.nrow())));
#endif
  }
  
  OUparam getParams(const MyMatrix& origParams,
		    bool& fail) const
  {
    OUparam newParams;
    
    fail = false;
    
#ifndef MONOCAR_NO_DEBUG
    assert(origParams.n_cols == 1);
#endif
    
    MyMatrix params(origParams);
    
    params.resize(params.n_rows + 1, 1);
    params(origParams.n_rows, 0) = 1.0;

    newParams.varByHouse = varByHouse;
    newParams.varCenters = varCenters;
    
    newParams.mean = meanConstraint * params;
    newParams.extraMean = extraMeanConstraint * params;

    newParams.extraVarAdd = extraVarAddConstraint * params;
    newParams.extraVarMult = extraVarMultConstraint * params;
    newParams.extraVarPow = extraVarPowConstraint * params;

    newParams.reversion = reversionConstraint * params;
    newParams.reversion.reshape((uint) std::sqrt(newParams.reversion.n_rows),
				(uint) std::sqrt(newParams.reversion.n_rows));


    if (newParams.mean.n_rows == newParams.reversion.n_rows * processIndexSize)
      newParams.mean.reshape((uint) newParams.reversion.n_rows, (uint) processIndexSize);
    if (newParams.extraMean.n_rows == ((uint) houseSize) * ((uint) processIndexSize))
      newParams.extraMean.reshape((uint) houseSize, (uint) processIndexSize);

    if (hasStart)
      newParams.start = (startConstraint * params);
    else
      newParams.start = (meanConstraint * params);
    newParams.hasStart = hasStart;
    
    newParams.hasExo = hasExo;
    if ((hasExo) && (exoConstraint.n_rows > 0))
      {
	newParams.exo = exoConstraint * params;
	newParams.hasExo = true;
      }
    else
      {
	newParams.hasExo = false;
      }

    MyMatrix sigmaSqrtUpper = sigmaSqrtConstraint * params;
    MyMatrix sigmaSqrt(newParams.reversion.n_rows, newParams.reversion.n_cols, fill::zeros);      

    uint pos = 0;
    for (uint j = 0; j < sigmaSqrt.n_rows; j++)
      {
	for (uint i = 0; i <= j; i++)
	  {
	    sigmaSqrt(j,i) = sigmaSqrtUpper(pos, 0);
	    pos++;
	  }
      }

    MyMatrix simultaneous = simultaneousConstraint * params;
    simultaneous.reshape((uint) std::sqrt(simultaneous.n_rows),
			 (uint) std::sqrt(simultaneous.n_rows));

    newParams.reversion = solve(simultaneous, newParams.reversion);
    sigmaSqrt = solve(simultaneous, sigmaSqrt);
    newParams.sigma = sigmaSqrt * sigmaSqrt.t();

    if (hasNegEigen(newParams.reversion)) {
      if (debugLevel > 2)
	Rcpp::Rcout << endl << "Matrix not valid:" << newParams.reversion << endl;
      else if (debugLevel == 2)
	Rcpp::Rcout << "Matrix not valid" << endl;
      fail = true;
    }

    return newParams;
  }



  OUparam getDParams(const MyMatrix& origParams,
		     const uint i) const
  {
    OUparam newParams;

    MyMatrix params(origParams.n_rows + 1, 1, fill::zeros);
    params(i, 0) = 1.0;

    MyMatrix realParams = origParams;
    realParams.resize(origParams.n_rows + 1, 1);
    realParams(origParams.n_rows, 0) = 1.0;

    newParams.varByHouse = varByHouse;
    newParams.varCenters = varCenters;

    newParams.mean = meanConstraint * params;
    newParams.extraMean = extraMeanConstraint * params;

    newParams.extraVarAdd = extraVarAddConstraint * params;
    newParams.extraVarMult = extraVarMultConstraint * params;
    newParams.extraVarPow = extraVarPowConstraint * params;

    newParams.reversion = reversionConstraint * params;
    newParams.reversion.reshape((uint) std::sqrt(newParams.reversion.n_rows),
				(uint) std::sqrt(newParams.reversion.n_rows));


    if (newParams.mean.n_rows == newParams.reversion.n_rows * processIndexSize)
      newParams.mean.reshape((uint) newParams.reversion.n_rows, (uint) processIndexSize);
    if (newParams.extraMean.n_rows == ((uint) houseSize) * ((uint) processIndexSize))
      newParams.extraMean.reshape((uint) houseSize, (uint) processIndexSize);

    if (hasStart)
      newParams.start = (startConstraint * params);
    else
      newParams.start = (meanConstraint * params);
    newParams.hasStart = hasStart;
    
    newParams.hasExo = hasExo;
    if ((hasExo) && (exoConstraint.n_rows > 0))
      {
	newParams.exo = exoConstraint * params;
	newParams.hasExo = true;
      }
    else
      {
	newParams.hasExo = false;
      }

    MyMatrix sigmaSqrtUpper = sigmaSqrtConstraint * params;
    MyMatrix sigmaSqrt(newParams.reversion.n_rows, newParams.reversion.n_cols, fill::zeros);      

    MyMatrix realSigmaSqrtUpper = sigmaSqrtConstraint * realParams;
    MyMatrix realSigmaSqrt(newParams.reversion.n_rows, newParams.reversion.n_cols, fill::zeros);      

    uint pos = 0;
    for (uint j = 0; j < sigmaSqrt.n_rows; j++)
      {
	for (uint k = 0; k <= j; k++)
	  {
	    sigmaSqrt(j,k) = sigmaSqrtUpper(pos, 0);
	    realSigmaSqrt(j,k) = realSigmaSqrtUpper(pos, 0);
	    pos++;
	  }
      }

    MyMatrix simultaneous = simultaneousConstraint * params;
    simultaneous.reshape((uint) std::sqrt(simultaneous.n_rows),
			 (uint) std::sqrt(simultaneous.n_rows));

    MyMatrix realSimultaneous = simultaneousConstraint * realParams;
    realSimultaneous.reshape((uint) std::sqrt(realSimultaneous.n_rows),
			     (uint) std::sqrt(realSimultaneous.n_rows));
    MyMatrix solveRealSimultaneous = inv(realSimultaneous);
    MyMatrix DsolveSimultaneous = - solveRealSimultaneous * simultaneous * solveRealSimultaneous;

    MyMatrix realReversion = reversionConstraint * realParams;
    realReversion.reshape((uint) std::sqrt(realReversion.n_rows),
			  (uint) std::sqrt(realReversion.n_rows));

    newParams.reversion = DsolveSimultaneous * realReversion + solveRealSimultaneous * newParams.reversion;
    sigmaSqrt = DsolveSimultaneous * realSigmaSqrt + solveRealSimultaneous * sigmaSqrt;
    realSigmaSqrt = solve(realSimultaneous, realSigmaSqrt);
    newParams.sigma = sigmaSqrt * realSigmaSqrt.t() + realSigmaSqrt * sigmaSqrt.t();

    return newParams;

  }

  OUparam getDDParams(const MyMatrix& origParams,
		      const uint i1,
		      const uint i2) const
  {
    OUparam newParams;

    MyMatrix params(origParams.n_rows + 1, 1, fill::zeros);

    MyMatrix params1(origParams.n_rows + 1, 1, fill::zeros);
    params1(i1, 0) = 1.0;
    MyMatrix params2(origParams.n_rows + 1, 1, fill::zeros);
    params2(i2, 0) = 1.0;
    
    MyMatrix realParams = origParams;
    realParams.resize(origParams.n_rows + 1, 1);
    realParams(origParams.n_rows, 0) = 1.0;

    newParams.varByHouse = varByHouse;
    newParams.varCenters = varCenters;

    newParams.mean = meanConstraint * params;
    newParams.extraMean = extraMeanConstraint * params;

    newParams.extraVarAdd = extraVarAddConstraint * params;
    newParams.extraVarMult = extraVarMultConstraint * params;
    newParams.extraVarPow = extraVarPowConstraint * params;

    newParams.reversion = reversionConstraint * params;
    newParams.reversion.reshape((uint) std::sqrt(newParams.reversion.n_rows),
				(uint) std::sqrt(newParams.reversion.n_rows));

    MyMatrix DDreversion = reversionConstraint * params;
    DDreversion.reshape((uint) std::sqrt(DDreversion.n_rows),
			(uint) std::sqrt(DDreversion.n_rows));

    MyMatrix D1reversion = reversionConstraint * params1;
    D1reversion.reshape((uint) std::sqrt(D1reversion.n_rows),
			(uint) std::sqrt(D1reversion.n_rows));
    
    MyMatrix D2reversion = reversionConstraint * params2;
    D2reversion.reshape((uint) std::sqrt(D2reversion.n_rows),
			(uint) std::sqrt(D2reversion.n_rows));

    MyMatrix realReversion = reversionConstraint * realParams;
    realReversion.reshape((uint) std::sqrt(realReversion.n_rows),
    			  (uint) std::sqrt(realReversion.n_rows));

    if (newParams.mean.n_rows == DDreversion.n_rows * processIndexSize)
      newParams.mean.reshape((uint) DDreversion.n_rows, (uint) processIndexSize);
    if (newParams.extraMean.n_rows == ((uint) houseSize) * ((uint) processIndexSize))
      newParams.extraMean.reshape((uint) houseSize, (uint) processIndexSize);

    if (hasStart)
      newParams.start = (startConstraint * params);
    else
      newParams.start = (meanConstraint * params);
    newParams.hasStart = hasStart;
    
    newParams.hasExo = hasExo;
    if ((hasExo) && (exoConstraint.n_rows > 0))
      {
	newParams.exo = exoConstraint * params;
	newParams.hasExo = true;
      }
    else
      {
	newParams.hasExo = false;
      }

    MyMatrix DDsigmaSqrtUpper = sigmaSqrtConstraint * params;
    MyMatrix DDsigmaSqrt(DDreversion.n_rows, DDreversion.n_cols, fill::zeros);

    MyMatrix D1sigmaSqrtUpper = sigmaSqrtConstraint * params1;
    MyMatrix D1sigmaSqrt(DDreversion.n_rows, DDreversion.n_cols, fill::zeros);

    MyMatrix D2sigmaSqrtUpper = sigmaSqrtConstraint * params2;
    MyMatrix D2sigmaSqrt(DDreversion.n_rows, DDreversion.n_cols, fill::zeros);
    
    MyMatrix realSigmaSqrtUpper = sigmaSqrtConstraint * realParams;
    MyMatrix realSigmaSqrt(DDreversion.n_rows, DDreversion.n_cols, fill::zeros);

    uint pos = 0;
    for (uint j = 0; j < DDsigmaSqrt.n_rows; j++)
      {
	for (uint i = 0; i <= j; i++)
	  {
	    DDsigmaSqrt(j,i) = DDsigmaSqrtUpper(pos, 0);
	    D1sigmaSqrt(j,i) = D1sigmaSqrtUpper(pos, 0);
	    D2sigmaSqrt(j,i) = D2sigmaSqrtUpper(pos, 0);
	    realSigmaSqrt(j,i) = realSigmaSqrtUpper(pos, 0);
	    pos++;
	  }
      }

    MyMatrix DDsimultaneous = simultaneousConstraint * params;
    DDsimultaneous.reshape((uint) std::sqrt(DDsimultaneous.n_rows),
    			 (uint) std::sqrt(DDsimultaneous.n_rows));

    MyMatrix D1simultaneous = simultaneousConstraint * params1;
    D1simultaneous.reshape((uint) std::sqrt(D1simultaneous.n_rows),
    			 (uint) std::sqrt(D1simultaneous.n_rows));

    MyMatrix D2simultaneous = simultaneousConstraint * params2;
    D2simultaneous.reshape((uint) std::sqrt(D2simultaneous.n_rows),
    			 (uint) std::sqrt(D2simultaneous.n_rows));

    MyMatrix realSimultaneous = simultaneousConstraint * realParams;
    realSimultaneous.reshape((uint) std::sqrt(realSimultaneous.n_rows),
    			     (uint) std::sqrt(realSimultaneous.n_rows));
    MyMatrix solveRealSimultaneous = inv(realSimultaneous);
    MyMatrix D1solveSimultaneous = - solveRealSimultaneous * D1simultaneous * solveRealSimultaneous;
    MyMatrix D2solveSimultaneous = - solveRealSimultaneous * D2simultaneous * solveRealSimultaneous;
    MyMatrix DDsolveSimultaneous =
      + solveRealSimultaneous * D2simultaneous * solveRealSimultaneous * D1simultaneous * solveRealSimultaneous
      - solveRealSimultaneous * DDsimultaneous * solveRealSimultaneous
      + solveRealSimultaneous * D1simultaneous * solveRealSimultaneous * D2simultaneous * solveRealSimultaneous;

    DDreversion =
      + DDsolveSimultaneous * realReversion
      + D1solveSimultaneous * D2reversion
      + D2solveSimultaneous * D1reversion
      + solveRealSimultaneous * DDreversion;
    
    DDsigmaSqrt =
      + DDsolveSimultaneous * realSigmaSqrt
      + D1solveSimultaneous * D2sigmaSqrt
      + D2solveSimultaneous * D1sigmaSqrt
      + solveRealSimultaneous * DDsigmaSqrt;
    D1sigmaSqrt = D1solveSimultaneous * realSigmaSqrt + solveRealSimultaneous * D1sigmaSqrt;
    D2sigmaSqrt = D2solveSimultaneous * realSigmaSqrt + solveRealSimultaneous * D2sigmaSqrt;
    realSigmaSqrt = solveRealSimultaneous * realSigmaSqrt;
    
    //D1sigmaSqrt = D1solveSimultaneous * realSigmaSqrt + solve(realSimultaneous, D1sigmaSqrt);
    //D2sigmaSqrt = D2solveSimultaneous * realSigmaSqrt + solve(realSimultaneous, D2sigmaSqrt);
    //realSigmaSqrt = solve(realSimultaneous, realSigmaSqrt);

    newParams.reversion = DDreversion;
    
    newParams.sigma =
      + DDsigmaSqrt * realSigmaSqrt.t()
      + D1sigmaSqrt * D2sigmaSqrt.t()
      + D2sigmaSqrt * D1sigmaSqrt.t()
      + realSigmaSqrt * DDsigmaSqrt.t();

    return newParams;

  }

private:
  const int processIndexSize;
  const int houseSize;
  const bool hasStart;
  const bool hasExo;
  const bool varByHouse;
  const MyMatrixCol varCenters;
  const MyMatrix startConstraint;
  const MyMatrix meanConstraint;
  const MyMatrix extraMeanConstraint;
  const MyMatrix reversionConstraint;
  const MyMatrix simultaneousConstraint;
  const MyMatrix sigmaSqrtConstraint;
  const MyMatrix extraVarAddConstraint;
  const MyMatrix extraVarMultConstraint;
  const MyMatrix extraVarPowConstraint;
  const MyMatrix exoConstraint;
  
  const int paramSize;
};

class OUprocess
{
public:
  OUprocess(const OUparam& iParam,
	    const vector<OUparam>& iDparam,
	    const vector<OUparam>& iDDparam,
	    const double timeZero,
	    bool iUseLong = false,
	    const bool iPartialScores = false)
    :partialScores(iPartialScores)
    ,logLik(0.0)
    ,seriesSize(iParam.mean.n_rows)
    ,params(iParam)
    ,Dparams(iDparam)
    ,DDparams(iDDparam)
    ,mu(iParam.mean - iParam.mean)
    ,P(iParam.reversion)
    ,Sigma(iParam.sigma)
    ,Ident(eye(iParam.reversion.n_rows, iParam.reversion.n_rows))
    ,PpP(kron(iParam.reversion,eye(iParam.reversion.n_rows, iParam.reversion.n_rows)) + kron(eye(iParam.reversion.n_rows, iParam.reversion.n_rows),iParam.reversion))
    ,sPpP(iParam.reversion.n_rows * iParam.reversion.n_rows, iParam.reversion.n_rows * iParam.reversion.n_rows)
    ,sP(inv(iParam.reversion))
    ,expP(iParam.reversion)
#ifdef START_NULL
    ,covHist(0, 0, fill::zeros)
    ,covHC(0, iParam.reversion.n_rows, fill::zeros)
#else
    ,covHist(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros)
    ,covHC(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros)
#endif
    ,covCurrent(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros)
    ,DlogLik(iDparam.size(), partialScores ? 0 : 1, fill::zeros)
    ,DexpP(iParam.reversion)
    ,DDlogLik(iDparam.size(), iDparam.size(), fill::zeros)
    ,DDexpP(iParam.reversion)
    ,lastTime(numeric_limits<double>::infinity())
#ifdef USE_LONG
    ,useLong(iUseLong)
#endif
    ,hasHess((iDDparam.size() > 0) && (iDDparam.size() == (iDparam.size() * (iDparam.size()+1) / 2)))
  {
    sPpP = inv(PpP);

#ifdef START_NULL
    meanHist = MyMatrix(0,1);
#else
    meanHist = iParam.start - iParam.mean;
#endif

    meanCurrent = iParam.start - iParam.mean;

    if (!iParam.hasStart)
      {
	MyMatrix cov33 =  sPpP * vectorise(Sigma);
	cov33.reshape(P.n_rows, P.n_cols);
	covCurrent = cov33;
      }

    for (vector<OUparam>::const_iterator i = iDparam.begin(); i != iDparam.end(); ++i)
      {
	DexpP.push_back(i->reversion);
	
#ifdef START_NULL
	DcovHist.push_back(MyMatrix(0, 0));
	DcovHC.push_back(MyMatrix(0, iParam.reversion.n_rows));
#else
	DcovHist.push_back(MyMatrix(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros));
	DcovHC.push_back(MyMatrix(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros));
#endif
	MyMatrix Dcov33(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros);
	
	MyMatrix DP = i->reversion;
	MyMatrix DSigma = i->sigma;
	MyMatrix DStart = (i->start) - (i->mean);

	DPpP.push_back(kron(DP,Ident) + kron(Ident,DP));
	DsPpP.push_back(- sPpP * DPpP.back() * sPpP);
	DsP.push_back(- sP * DP * sP);
	Dmu.push_back(i->mean - i->mean);
	
#ifdef START_NULL
	DmeanHist.push_back(MyMatrix(0,DStart.n_cols));
#else
	DmeanHist.push_back(DStart);
#endif
	DmeanCurrent.push_back(DStart);

	if (!iParam.hasStart)
	  {
	    Dcov33 =  DsPpP.back() * vectorise(Sigma) + sPpP * vectorise(DSigma);
	    Dcov33.reshape(P.n_rows, P.n_cols);
	  }

	DcovCurrent.push_back(Dcov33);

      }

    if (hasHess)
      {
	uint k = 0;
	uint icount = 0;
	for (vector<OUparam>::const_iterator i = iDparam.begin(); i != iDparam.end(); ++i)
	  {
	    icount++;
	    uint jcount = icount;

	    MyMatrix D1P = i->reversion;
	    MyMatrix D1Sigma = i->sigma;
	    MyMatrix D1Start = (i->start) - (i->mean);
	    MyMatrix D1PpP = kron(D1P,Ident) + kron(Ident,D1P);
	    MyMatrix D1sPpP = - sPpP * D1PpP * sPpP;

	    for (vector<OUparam>::const_iterator j = i; j != iDparam.end(); ++j)
	      {
		jcount++;
		DDexpP.push_back(i->reversion, j->reversion, iDDparam[k].reversion);
		
#ifdef START_NULL
		DDcovHist.push_back(MyMatrix(0, 0));
		DDcovHC.push_back(MyMatrix(0, iParam.reversion.n_rows));
#else
		DDcovHist.push_back(MyMatrix(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros));
		DDcovHC.push_back(MyMatrix(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros));
#endif
		MyMatrix DDcov33(iParam.reversion.n_rows, iParam.reversion.n_rows, fill::zeros);
		
		MyMatrix D2P = j->reversion;
		MyMatrix D1D2P = iDDparam[k].reversion;
		MyMatrix D2Sigma = j->sigma;
		MyMatrix D1D2Sigma = iDDparam[k].sigma;
		MyMatrix D2Start = (j->start) - (j->mean);
		MyMatrix DDStart = iDDparam[k].start - iDDparam[k].mean;;
		
		DDPpP.push_back(kron(D1D2P,Ident) + kron(Ident,D1D2P));
		MyMatrix D2PpP = kron(D2P,Ident) + kron(Ident,D2P);
		DDsPpP.push_back(sPpP * D2PpP * sPpP * D1PpP * sPpP
				 - sPpP * DDPpP.back() * sPpP
				 + sPpP * D1PpP * sPpP * D2PpP * sPpP);
		MyMatrix D2sPpP = - sPpP * D2PpP * sPpP;
		DDsP.push_back(+ sP * D2P * sP * D1P * sP
			      - sP * D1D2P * sP
			      + sP * D1P * sP * D2P * sP);
		DDmu.push_back(iDDparam[k].mean - iDDparam[k].mean);
		
#ifdef START_NULL
		DDmeanHist.push_back(MyMatrix(0,DDStart.n_cols));
#else
		DDmeanHist.push_back(DDStart);
#endif
		DDmeanCurrent.push_back(DDStart);
		
		if (!iParam.hasStart)
		  {
		    DDcov33 =
		      + DDsPpP.back() * vectorise(Sigma)
		      + D1sPpP * vectorise(D2Sigma)
		      + D2sPpP * vectorise(D1Sigma)
		      + sPpP * vectorise(D1D2Sigma);
		    DDcov33.reshape(P.n_rows, P.n_cols);
		  }
		
		DDcovCurrent.push_back(DDcov33);
		
		k++;
	      }
	  }
      }


#ifdef USE_LONG
    if (useLong)
      {
#ifdef START_NULL
	meanLong = meanHist;
	covLong = covHist;
	covLC = covHC;
	covLongDiag = MyMatrix(0,0);	
#else
	meanLong = 0*meanCurrent;
	covLong = 0*P;
	covLC = 0*P;
	covLongDiag = 0*meanCurrent;
#endif
      }
#endif

#ifndef MONOCAR_NO_DEBUG
    assert(iP.is_square());
    assert(iSigma.is_square());
#endif

    timeIndices.clear();
#ifndef START_NULL
    timeIndices.push_back(timeZero - 10.0);
#endif
    timeIndices.push_back(timeZero - 0.0);

    fullTimeIndices.clear();
#ifdef START_NULL
    fullTimeIndices.push_back(timeZero - 10.0);
#endif
    fullTimeIndices.push_back(timeZero - 0.0);
    
  }

  MyMatrixCol sampleCurrent(void)
  {
    MyMatrixCol obs = rmvnorm(meanCurrent, covCurrent);
    meanCurrent = obs;
    meanHist += covHC * solve(covCurrent, obs - meanCurrent);
    covHist -= covHC * solve(covCurrent, covHC.t());
    covCurrent.zeros();
    covHC.zeros();
    return(obs);
  }

  MyMatrixCol sampleHist(void)
  {
    MyMatrixCol obs = rmvnorm(meanHist, covHist);
    meanHist = obs;
    meanCurrent += covHC.t() * solve(covHist, obs - meanHist);
    covCurrent -= covHC.t() * solve(covHist, covHC);
    covHist.zeros();
    covHC.zeros();
    return(obs);
  }
  
  void addTime(const double timeEnd)
  {
#ifndef MONOCAR_NO_DEBUG
    assert(timeEnd > timeIndices[timeIndices.size()-1]);
#endif
    
    double time = timeEnd - timeIndices[timeIndices.size()-1];

    timeIndices.push_back(timeEnd);

    if (time != lastTime)
      {
	expPnegtime = expP(-time);
	IexpPpPnegtime = -kron(expPnegtime,expPnegtime);
	IexpPpPnegtime.diag() += 1.0;
	lastTime = time;
      }

    MyMatrix PinvIexpPntot = (sP * (Ident-expPnegtime)) / time;

    
    MyMatrix cov33 = sPpP * (IexpPpPnegtime) * vectorise(Sigma);
    cov33.reshape(P.n_rows, P.n_cols);


    MyMatrix cov12 = covHC * PinvIexpPntot.t();


    MyMatrix cov22 = sP * (Ident - PinvIexpPntot) * Sigma
      - PinvIexpPntot * Sigma * sP.t()
      + cov33 * sP.t() / time;
    cov22.reshape(P.n_rows * P.n_cols, 1);

    MyMatrix cov23 = Sigma * PinvIexpPntot.t()
      - PinvIexpPntot * Sigma * expPnegtime.t();
    cov23.reshape(P.n_rows*P.n_cols, 1);


    if (hasHess)
      {
	uint k = 0;
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    MyMatrix D1expPnegtime = DexpP(-time, i);
	    MyMatrix D1IexpPpPnegtime =
	      -kron(D1expPnegtime,expPnegtime)
	      -kron(expPnegtime,D1expPnegtime);
	    MyMatrix D1PinvIexpPntot = (DsP[i] * (Ident-expPnegtime) - sP * D1expPnegtime) / time;
	    MyMatrix D1Sigma = Dparams[i].sigma;
	    
	    MyMatrix D1cov33 =
	      DsPpP[i] * (IexpPpPnegtime) * vectorise(Sigma)
	      + sPpP * (D1IexpPpPnegtime) * vectorise(Sigma)
	      + sPpP * (IexpPpPnegtime) * vectorise(D1Sigma);
	    D1cov33.reshape(P.n_rows, P.n_cols);
		
	    MyMatrix D1cov22 = DsP[i] * (Ident - PinvIexpPntot) * Sigma
	      - sP * (D1PinvIexpPntot) * Sigma
	      + sP * (Ident - PinvIexpPntot) * D1Sigma
	      - D1PinvIexpPntot * Sigma * sP.t()
	      - PinvIexpPntot * D1Sigma * sP.t()
	      - PinvIexpPntot * Sigma * DsP[i].t()
	      + D1cov33 * sP.t() / time
	      + cov33 * DsP[i].t() / time;
	    D1cov22.reshape(P.n_rows * P.n_cols, 1);
	    // D1cov22 = (DsPpP[i] * cov22 + sPpP * D1cov22) / time;
	    // D1cov22.reshape(P.n_rows, P.n_cols);
	    // D1cov22 += D1cov22.t();
		
	    MyMatrix D1cov23 =
	      D1Sigma * PinvIexpPntot.t()
	      + Sigma * D1PinvIexpPntot.t()
	      - D1PinvIexpPntot * Sigma * expPnegtime.t()
	      - PinvIexpPntot * D1Sigma * expPnegtime.t()
	      - PinvIexpPntot * Sigma * D1expPnegtime.t();
	    D1cov23.reshape(P.n_rows*P.n_cols, 1);
	    // D1cov23 = DsPpP[i] * cov23 + sPpP * D1cov23;
	    // D1cov23.reshape(P.n_rows, P.n_cols);

	    for (uint j = i; j < Dparams.size(); j++)
	      {
		
		
		MyMatrix D2expPnegtime = DexpP(-time, j);
		MyMatrix D2IexpPpPnegtime =
		  -kron(D2expPnegtime,expPnegtime)
		  -kron(expPnegtime,D2expPnegtime);
		MyMatrix PinvIexpPntot = (sP * (Ident-expPnegtime)) / time;
		MyMatrix D2PinvIexpPntot = (DsP[j] * (Ident-expPnegtime) - sP * D2expPnegtime) / time;
		MyMatrix D2Sigma = Dparams[j].sigma;
		
		MyMatrix D1D2expPnegtime = DDexpP(-time, k);
		MyMatrix D1D2IexpPpPnegtime =
		  -kron(D1D2expPnegtime,expPnegtime)
		  -kron(D1expPnegtime,D2expPnegtime)
		  -kron(D2expPnegtime,D1expPnegtime)
		  -kron(expPnegtime,D1D2expPnegtime);
		MyMatrix D1D2PinvIexpPntot = (DDsP[k] * (Ident-expPnegtime)
					      - DsP[i] * D2expPnegtime
					      - DsP[j] * D1expPnegtime
					      - sP * D1D2expPnegtime) / time;
		MyMatrix D1D2Sigma = DDparams[k].sigma;

		MyMatrix D2cov33 =
		  DsPpP[j] * (IexpPpPnegtime) * vectorise(Sigma)
		  + sPpP * (D2IexpPpPnegtime) * vectorise(Sigma)
		  + sPpP * (IexpPpPnegtime) * vectorise(D2Sigma);
		D2cov33.reshape(P.n_rows, P.n_cols);
	    
		MyMatrix DDcov33 =
		  + DDsPpP[k] * (IexpPpPnegtime) * vectorise(Sigma)
		  + DsPpP[j] * (D1IexpPpPnegtime) * vectorise(Sigma)
		  + DsPpP[j] * (IexpPpPnegtime) * vectorise(D1Sigma)
		  + DsPpP[i] * (D2IexpPpPnegtime) * vectorise(Sigma)
		  + sPpP * (D1D2IexpPpPnegtime) * vectorise(Sigma)
		  + sPpP * (D2IexpPpPnegtime) * vectorise(D1Sigma)
		  + DsPpP[i] * (IexpPpPnegtime) * vectorise(D2Sigma)
		  + sPpP * (D1IexpPpPnegtime) * vectorise(D2Sigma)
		  + sPpP * (IexpPpPnegtime) * vectorise(D1D2Sigma);
		DDcov33.reshape(P.n_rows, P.n_cols);

		MyMatrix D2cov22 = DsP[j] * (Ident - PinvIexpPntot) * Sigma
		  - sP * (D2PinvIexpPntot) * Sigma
		  + sP * (Ident - PinvIexpPntot) * D2Sigma
		  - D2PinvIexpPntot * Sigma * sP.t()
		  - PinvIexpPntot * D2Sigma * sP.t()
		  - PinvIexpPntot * Sigma * DsP[j].t()
		  + D2cov33 * sP.t() / time
		  + cov33 * DsP[j].t() / time;
		D2cov22.reshape(P.n_rows * P.n_cols, 1);
		// D2cov22 = (DsPpP[j] * cov22 + sPpP * D2cov22) / time;
		// D2cov22.reshape(P.n_rows, P.n_cols);
		// D2cov22 += D2cov22.t();
	    
	    
		MyMatrix DDcov22 =
		  + DDsP[k] * (Ident - PinvIexpPntot) * Sigma
		  - DsP[i] * D2PinvIexpPntot * Sigma
		  + DsP[i] * (Ident - PinvIexpPntot) * D2Sigma
		  
		  - DsP[j] * (D1PinvIexpPntot) * Sigma
		  - sP * (D1D2PinvIexpPntot) * Sigma
		  - sP * (D1PinvIexpPntot) * D2Sigma
		  
		  + DsP[j] * (Ident - PinvIexpPntot) * D1Sigma
		  - sP * D2PinvIexpPntot * D1Sigma
		  + sP * (Ident - PinvIexpPntot) * D1D2Sigma
		  
		  - D1D2PinvIexpPntot * Sigma * sP.t()
		  - D1PinvIexpPntot * D2Sigma * sP.t()
		  - D1PinvIexpPntot * Sigma * DsP[j].t()
		  
		  - D2PinvIexpPntot * D1Sigma * sP.t()
		  - PinvIexpPntot * D1D2Sigma * sP.t()
		  - PinvIexpPntot * D1Sigma * DsP[j].t()
		  
		  - D2PinvIexpPntot * Sigma * DsP[i].t()
		  - PinvIexpPntot * D2Sigma * DsP[i].t()
		  - PinvIexpPntot * Sigma * DDsP[k].t()
	      
		  + DDcov33 * sP.t() / time
		  + D1cov33 * DsP[j].t() / time	      
		  + D2cov33 * DsP[i].t() / time
		  + cov33 * DDsP[k].t() / time;
	    
		DDcov22.reshape(P.n_rows * P.n_cols, 1);
		DDcov22 = (DDsPpP[k] * cov22 + DsPpP[i] * D2cov22 + DsPpP[j] * D1cov22 + sPpP * DDcov22) / time;
		DDcov22.reshape(P.n_rows, P.n_cols);
		DDcov22 += DDcov22.t();
				
		MyMatrix D2cov23 =
		  D2Sigma * PinvIexpPntot.t()
		  + Sigma * D2PinvIexpPntot.t()
		  - D2PinvIexpPntot * Sigma * expPnegtime.t()
		  - PinvIexpPntot * D2Sigma * expPnegtime.t()
		  - PinvIexpPntot * Sigma * D2expPnegtime.t();
		D2cov23.reshape(P.n_rows*P.n_cols, 1);
		// D2cov23 = DsPpP[j] * cov23 + sPpP * D2cov23;
		// D2cov23.reshape(P.n_rows, P.n_cols);
		
		MyMatrix DDcov23 =
		  + D1D2Sigma * PinvIexpPntot.t()
		  + D1Sigma * D2PinvIexpPntot.t()
		  
		  + D2Sigma * D1PinvIexpPntot.t()
		  + Sigma * D1D2PinvIexpPntot.t()
		  
		  - D1D2PinvIexpPntot * Sigma * expPnegtime.t()
		  - D1PinvIexpPntot * D2Sigma * expPnegtime.t()
		  - D1PinvIexpPntot * Sigma * D2expPnegtime.t()
		  
		  - D2PinvIexpPntot * D1Sigma * expPnegtime.t()
		  - PinvIexpPntot * D1D2Sigma * expPnegtime.t()
		  - PinvIexpPntot * D1Sigma * D2expPnegtime.t()
		  
		  - D2PinvIexpPntot * Sigma * D1expPnegtime.t()
		  - PinvIexpPntot * D2Sigma * D1expPnegtime.t()
		  - PinvIexpPntot * Sigma * D1D2expPnegtime.t();
		
		DDcov23.reshape(P.n_rows*P.n_cols, 1);
		DDcov23 = DDsPpP[k] * cov23 + DsPpP[i] * D2cov23 + DsPpP[j] * D1cov23 + sPpP * DDcov23;
		DDcov23.reshape(P.n_rows, P.n_cols);
	    
		// MyMatrix D1cov12 = DcovHC[i] * PinvIexpPntot.t() + covHC * D1PinvIexpPntot.t();
		// MyMatrix D2cov12 = DcovHC[j] * PinvIexpPntot.t() + covHC * D2PinvIexpPntot.t();	    
		MyMatrix DDcov12 =
		  + DDcovHC[k] * PinvIexpPntot.t()
		  + DcovHC[i] * D2PinvIexpPntot.t()
		  + DcovHC[j] * D1PinvIexpPntot.t()
		  + covHC * D1D2PinvIexpPntot.t();

		DDmeanHist[k] = join_cols(DDmeanHist[k],
					  + D1D2PinvIexpPntot * (meanCurrent - mu)
					  + D1PinvIexpPntot * (DmeanCurrent[j] - Dmu[j])
					  + D2PinvIexpPntot * (DmeanCurrent[i] - Dmu[i])
					  + PinvIexpPntot * (DDmeanCurrent[k] - DDmu[k]));
		DDmeanCurrent[k] =
		  + D1D2expPnegtime * (meanCurrent - mu)
		  + D1expPnegtime * (DmeanCurrent[j] - Dmu[j])
		  + D2expPnegtime * (DmeanCurrent[i] - Dmu[i])
		  + expPnegtime * (DDmeanCurrent[k] - DDmu[k]);
		
		DDcovHC[k] = join_cols(DDcovHC[k] * expPnegtime.t()
				       + DcovHC[i] * D2expPnegtime.t()
				       + DcovHC[j] * D1expPnegtime.t()
				       + covHC * D1D2expPnegtime.t(),
				       DDcov23
				       
				       + D1D2PinvIexpPntot * covCurrent * expPnegtime.t()
				       + D1PinvIexpPntot * DcovCurrent[j] * expPnegtime.t()
				       + D1PinvIexpPntot * covCurrent * D2expPnegtime.t()
				       
				       + D2PinvIexpPntot * DcovCurrent[i] * expPnegtime.t()
				       + PinvIexpPntot * DDcovCurrent[k] * expPnegtime.t()
				       + PinvIexpPntot * DcovCurrent[i] * D2expPnegtime.t()
				   
				       + D2PinvIexpPntot * covCurrent * D1expPnegtime.t()
				       + PinvIexpPntot * DcovCurrent[j] * D1expPnegtime.t()
				       + PinvIexpPntot * covCurrent * D1D2expPnegtime.t()
				       );
	    
		DDcovHist[k] = join_cols(join_rows(DDcovHist[k], DDcov12),
					 join_rows(DDcov12.t(),
						   DDcov22

						   + D1D2PinvIexpPntot * covCurrent * PinvIexpPntot.t()
						   + D1PinvIexpPntot * DcovCurrent[j] * PinvIexpPntot.t()
						   + D1PinvIexpPntot * covCurrent * D2PinvIexpPntot.t()
						   
						   + D2PinvIexpPntot * DcovCurrent[i] * PinvIexpPntot.t()
						   + PinvIexpPntot * DDcovCurrent[k] * PinvIexpPntot.t()
						   + PinvIexpPntot * DcovCurrent[i] * D2PinvIexpPntot.t()
					       
						   + D2PinvIexpPntot * covCurrent * D1PinvIexpPntot.t()
						   + PinvIexpPntot * DcovCurrent[j] * D1PinvIexpPntot.t()
						   + PinvIexpPntot * covCurrent * D1D2PinvIexpPntot.t()
						   ));
	    
		DDcovCurrent[k] = DDcov33

		  + D1D2expPnegtime * covCurrent * expPnegtime.t()
		  + D1expPnegtime * DcovCurrent[j] * expPnegtime.t()
		  + D1expPnegtime * covCurrent * D2expPnegtime.t()
		  
		  + D2expPnegtime * DcovCurrent[i] * expPnegtime.t()
		  + expPnegtime * DDcovCurrent[k] * expPnegtime.t()
		  + expPnegtime * DcovCurrent[i] * D2expPnegtime.t()
		  
		  + D2expPnegtime * covCurrent * D1expPnegtime.t()
		  + expPnegtime * DcovCurrent[j] * D1expPnegtime.t()
		  + expPnegtime * covCurrent * D1D2expPnegtime.t();
		  
		k++;
	      }
	  }
      }


    for (uint i = 0; i < Dparams.size(); i++)
      {

	MyMatrix DexpPnegtime = DexpP(-time, i);

	MyMatrix DIexpPpPnegtime = -kron(DexpPnegtime,expPnegtime) -kron(expPnegtime,DexpPnegtime);
	MyMatrix DPinvIexpPntot = (DsP[i] * (Ident-expPnegtime) - sP * DexpPnegtime) / time;
	MyMatrix DSigma = Dparams[i].sigma;
		
	MyMatrix Dcov33 =
	  DsPpP[i] * (IexpPpPnegtime) * vectorise(Sigma)
	  + sPpP * (DIexpPpPnegtime) * vectorise(Sigma)
	  + sPpP * (IexpPpPnegtime) * vectorise(DSigma);
	Dcov33.reshape(P.n_rows, P.n_cols);

	MyMatrix Dcov22 = DsP[i] * (Ident - PinvIexpPntot) * Sigma
	  - sP * (DPinvIexpPntot) * Sigma
	  + sP * (Ident - PinvIexpPntot) * DSigma
	  - DPinvIexpPntot * Sigma * sP.t()
	  - PinvIexpPntot * DSigma * sP.t()
	  - PinvIexpPntot * Sigma * DsP[i].t()
	  + Dcov33 * sP.t() / time
	  + cov33 * DsP[i].t() / time;
	Dcov22.reshape(P.n_rows * P.n_cols, 1);
	Dcov22 = (DsPpP[i] * cov22 + sPpP * Dcov22) / time;
	Dcov22.reshape(P.n_rows, P.n_cols);
	Dcov22 += Dcov22.t();

	MyMatrix Dcov23 =
	  DSigma * PinvIexpPntot.t()
	  + Sigma * DPinvIexpPntot.t()
	  - DPinvIexpPntot * Sigma * expPnegtime.t()
	  - PinvIexpPntot * DSigma * expPnegtime.t()
	  - PinvIexpPntot * Sigma * DexpPnegtime.t();
	Dcov23.reshape(P.n_rows*P.n_cols, 1);
	Dcov23 = DsPpP[i] * cov23 + sPpP * Dcov23;
	Dcov23.reshape(P.n_rows, P.n_cols);
    
	MyMatrix Dcov12 = DcovHC[i] * PinvIexpPntot.t() + covHC * DPinvIexpPntot.t();

	DmeanHist[i] = join_cols(DmeanHist[i], PinvIexpPntot * (DmeanCurrent[i] - Dmu[i]) + DPinvIexpPntot * (meanCurrent - mu) + Dmu[i]);
    
	DmeanCurrent[i] = expPnegtime * (DmeanCurrent[i] - Dmu[i]) + DexpPnegtime * (meanCurrent - mu) + Dmu[i];

	DcovHC[i] = join_cols(DcovHC[i] * expPnegtime.t()
			   + covHC * DexpPnegtime.t(),
			   Dcov23
			   + DPinvIexpPntot * covCurrent * expPnegtime.t()
			   + PinvIexpPntot * DcovCurrent[i] * expPnegtime.t()
			   + PinvIexpPntot * covCurrent * DexpPnegtime.t() );

	DcovHist[i] = join_cols(join_rows(DcovHist[i], Dcov12),
				join_rows(Dcov12.t(),
					  Dcov22
					  + DPinvIexpPntot * covCurrent * PinvIexpPntot.t()
					  + PinvIexpPntot * DcovCurrent[i] * PinvIexpPntot.t()
					  + PinvIexpPntot * covCurrent * DPinvIexpPntot.t()
					  ));

	DcovCurrent[i] = Dcov33
	  + DexpPnegtime * covCurrent * expPnegtime.t()
	  + expPnegtime * DcovCurrent[i] * expPnegtime.t()
	  + expPnegtime * covCurrent * DexpPnegtime.t();
	
      }


    
    cov22 = sPpP * cov22 / time;
    cov22.reshape(P.n_rows, P.n_cols);
    cov22 += cov22.t();

    cov23 = sPpP * cov23;
    cov23.reshape(P.n_rows, P.n_cols);


#ifdef USE_LONG
    if (useLong)
      {
	covLong = join_cols(covLong, PinvIexpPntot * covLC);
	covLC = expPnegtime * covLC;
      }
#endif

    meanHist = join_cols(meanHist, PinvIexpPntot * (meanCurrent - mu) + mu);

    meanCurrent = expPnegtime * (meanCurrent - mu) + mu;

    covHC = join_cols(covHC * expPnegtime.t(), PinvIexpPntot * covCurrent * expPnegtime.t() + cov23);
    
    covHist = join_cols(join_rows(covHist, cov12),
			join_rows(cov12.t(),
				  cov22
				  + PinvIexpPntot * covCurrent * PinvIexpPntot.t()
				  ));

    covCurrent = expPnegtime * covCurrent * expPnegtime.t() + cov33;

  }

  void addInstantaneousObservation(int series, double timeBegin, double obs, double var, const MyMatrixRow& exoDataRow, const int seriesIndex, const int houseIndex, const uint k)
  {
    if (timeIndices[timeIndices.size()-1] != timeBegin)
      {
	Rcpp::Rcout << "index["<<timeIndices.size()-1<<"]: " <<
	  timeIndices[timeIndices.size()-1] << endl;
	Rcpp::Rcout << "time: " << timeBegin << endl;
      }

#ifndef MONOCAR_NO_DEBUG
    assert(timeIndices[timeIndices.size()-1] == timeBegin);
#endif

    double varPow = 0;
    if ((houseIndex >= 0) & (var > 0))
      varPow = params.getVarCenter(houseIndex, seriesIndex) *
	std::pow(var/params.getVarCenter(houseIndex, seriesIndex),
		 params.getExtraVarPow(houseIndex, seriesIndex));


    vector<double> Dobs;
    vector<double> Dvar;

    for (uint i = 0; i < Dparams.size(); i++)
      {
	Dobs.push_back(0.0);
	Dvar.push_back(0.0);

	Dobs[i] -= Dparams[i].mean(seriesIndex, k < params.mean.n_cols ? k : 0);
	if (houseIndex >= 0)
	  Dobs[i] -= Dparams[i].extraMean(houseIndex, k < params.extraMean.n_cols ? k : 0);

	if (params.hasExo)
	  Dobs[i] -= as_scalar(exoDataRow * Dparams[i].exo);

	if (houseIndex >= 0)
	  {
	    if (var > 0)
	      Dvar[i] = params.getExtraVarMult(houseIndex, seriesIndex)
		* Dparams[i].getExtraVarPow(houseIndex, seriesIndex)
		* log(var/params.getVarCenter(houseIndex, seriesIndex))
		* varPow
		+ Dparams[i].getExtraVarMult(houseIndex, seriesIndex)
		* varPow
		+ Dparams[i].getExtraVarAdd(houseIndex, seriesIndex);
	    else
	      Dvar[i] = Dparams[i].getExtraVarAdd(houseIndex, seriesIndex);	      
	  }
      }


    vector<double> DDobs;
    vector<double> DDvar;

    if (hasHess)
      {
	uint ij = 0;
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    for (uint j = i; j < Dparams.size(); j++)
	      {
		DDobs.push_back(0.0);
		DDvar.push_back(0.0);
		
		DDobs[ij] -= DDparams[ij].mean(seriesIndex, k < params.mean.n_cols ? k : 0);
		if (houseIndex >= 0)
		  DDobs[ij] -= DDparams[ij].extraMean(houseIndex, k < params.extraMean.n_cols ? k : 0);
		
		if (params.hasExo)
		  DDobs[ij] -= as_scalar(exoDataRow * DDparams[ij].exo);
		
		if (houseIndex >= 0)
		  {
		    if (var > 0)
		      DDvar[ij] =
			+ Dparams[j].getExtraVarMult(houseIndex, seriesIndex)
			* Dparams[i].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
			
			+ params.getExtraVarMult(houseIndex, seriesIndex)
			* DDparams[ij].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		    
			+ params.getExtraVarMult(houseIndex, seriesIndex)
			* Dparams[i].getExtraVarPow(houseIndex, seriesIndex)
			* Dparams[j].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		
			+ DDparams[ij].getExtraVarMult(houseIndex, seriesIndex)
			* varPow
			
			+ Dparams[i].getExtraVarMult(houseIndex, seriesIndex)
			* Dparams[j].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		    
			+ DDparams[ij].getExtraVarAdd(houseIndex, seriesIndex);
		    else
		      DDvar[ij] = DDparams[ij].getExtraVarAdd(houseIndex, seriesIndex);
		    
		    ij++;
		  }
	      }
	  }
      }


    obs -= params.mean(seriesIndex, k < params.mean.n_cols ? k : 0);
    if (houseIndex >= 0)
      obs -= params.extraMean(houseIndex, k < params.extraMean.n_cols ? k : 0);

    if (params.hasExo)
      obs -= as_scalar(exoDataRow * params.exo);


    if (houseIndex >= 0)
      {
	var = params.getExtraVarMult(houseIndex, seriesIndex)
	  * varPow
	  + params.getExtraVarAdd(houseIndex, seriesIndex);
      }

    MyMatrix H(1, covCurrent.n_cols, fill::zeros);
    H(0,series) = 1.0;

    double y = obs - as_scalar(H * meanCurrent);
    double S = as_scalar(H * covCurrent * H.t()) + var;

    if (S > 0.0)
      {	
	MyMatrix K = covCurrent * H.t() / S;
	logLik -= (y * y / S + log(S) + LOGTWOPI) / 2.0;
	MyMatrix HcHC = H * covHC.t();


	if (hasHess)
	  {
	    uint ij = 0;
	    for (uint i = 0; i < Dparams.size(); i++)
	      {
		double D1y = Dobs[i] - as_scalar(H * DmeanCurrent[i]);
		double D1S = as_scalar(H * DcovCurrent[i] * H.t()) + Dvar[i];
		double D1Sinv = - D1S / (S*S);
		MyMatrix D1K = DcovCurrent[i] * H.t() / S + covCurrent * H.t() * D1Sinv;
		MyMatrix D1HcHC = H * DcovHC[i].t();

		for (uint j = i; j < Dparams.size(); j++)
		  {
		    double D2y = Dobs[j] - as_scalar(H * DmeanCurrent[j]);
		    double D2S = as_scalar(H * DcovCurrent[j] * H.t()) + Dvar[j];
		    double D2Sinv = - D2S / (S*S);
		    MyMatrix D2K = DcovCurrent[j] * H.t() / S + covCurrent * H.t() * D2Sinv;

		    double DDy = DDobs[ij] - as_scalar(H * DDmeanCurrent[ij]);
		    double DDS = as_scalar(H * DDcovCurrent[ij] * H.t()) + DDvar[ij];
		    double DDSinv = 2 * D1S * D2S / (S*S*S) - DDS / (S*S);
		    MyMatrix DDK =
		      + DDcovCurrent[ij] * H.t() / S
		      + DcovCurrent[i] * H.t() * D2Sinv
		      + DcovCurrent[j] * H.t() * D1Sinv
		      + covCurrent * H.t() * DDSinv;

		    double DDlogLikChange =
		      + DDy * y / S
		      + D1y * D2y / S
		      + D1y * y * D2Sinv
		      + D2y * y * D1Sinv
		      + (y * y * DDSinv) / 2.0
		      + (DDS / S) / 2.0
		      - (D1S * D2S / (S*S)) / 2.0;
		    
		    DDlogLik(i,j) -= DDlogLikChange;
		    if (i != j)
		      DDlogLik(j,i) -= DDlogLikChange;
		  
		    MyMatrix D2HcHC = H * DcovHC[j].t();
		    MyMatrix DDHcHC = H * DDcovHC[ij].t();
	    
		    DDmeanCurrent[ij] += DDK * y + D1K * D2y + D2K * D1y + K * DDy;
		  
		    DDmeanHist[ij] +=
		      + DDHcHC.t() * (y / S)
		      + D1HcHC.t() * (D2y / S)
		      + D1HcHC.t() * (y * D2Sinv)
		    
		      + D2HcHC.t() * (D1y / S)
		      + HcHC.t() * (DDy / S)
		      + HcHC.t() * (D1y * D2Sinv)
		    
		      + D2HcHC.t() * (y * D1Sinv)
		      + HcHC.t() * (D2y * D1Sinv)
		      + HcHC.t() * (y * DDSinv);
		  
		    DDcovCurrent[ij] -=
		      + DDK * H * covCurrent
		      + D1K * H * DcovCurrent[j]
		      + D2K * H * DcovCurrent[i]
		      + K * H * DDcovCurrent[ij];
		  
		    DDcovHist[ij] -=
		      + DDHcHC.t() * HcHC / S
		      + D1HcHC.t() * D2HcHC / S
		      + D1HcHC.t() * HcHC * D2Sinv
		    
		      + D2HcHC.t() * D1HcHC / S
		      + HcHC.t() * DDHcHC / S
		      + HcHC.t() * D1HcHC * D2Sinv
		    
		      + D2HcHC.t() * HcHC * D1Sinv
		      + HcHC.t() * D2HcHC * D1Sinv
		      + HcHC.t() * HcHC * DDSinv;
		  
		    DDcovHC[ij] -= trans(DDK * HcHC + D1K * D2HcHC + D2K * D1HcHC + K * DDHcHC);
		  
		    ij++;
		  }
	      }
	  }

	
	if (partialScores)
	  DlogLik.resize(DlogLik.n_rows, DlogLik.n_cols + 1);
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    double Dy = Dobs[i] - as_scalar(H * DmeanCurrent[i]);
	    double DS = as_scalar(H * DcovCurrent[i] * H.t()) + Dvar[i];
	    double DSinv = - DS / (S*S);
	    MyMatrix DK = DcovCurrent[i] * H.t() / S + covCurrent * H.t() * DSinv;

	    if (partialScores)
	      DlogLik(i,DlogLik.n_cols-1) = -(Dy * y / S + (y * y * DSinv + DS / S) / 2.0);
	    else
	      DlogLik(i,0) -= Dy * y / S + (y * y * DSinv + DS / S) / 2.0;

	    MyMatrix DHcHC = H * DcovHC[i].t();
	    
	    DmeanCurrent[i] += DK * y + K * Dy;
	    DmeanHist[i] += DHcHC.t() * (y / S) + HcHC.t() * (Dy / S) + HcHC.t() * (y * DSinv);
	    DcovCurrent[i] -= DK * H * covCurrent + K * H * DcovCurrent[i];
	    DcovHist[i] -= DHcHC.t() * HcHC / S + HcHC.t() * DHcHC / S + HcHC.t() * HcHC * DSinv;
	    DcovHC[i] -= trans(DK * HcHC + K * DHcHC);
	  }

	meanCurrent += K * y;
	meanHist += HcHC.t() * (y / S);
	covCurrent -= K * H * covCurrent;
	covHist -= HcHC.t() * HcHC / S;
	covHC -= trans(K * HcHC);

      }
    else
      {
	logLik -= y * y / 2.0;
	if (hasHess)
	  {
	    uint ij = 0;
	    for (uint i = 0; i < Dparams.size(); i++)
	      {
		double D1y = Dobs[i] - as_scalar(H * DmeanCurrent[i]);
		for (uint j = i; j < Dparams.size(); j++)
		  {
		    double D2y = Dobs[j] - as_scalar(H * DmeanCurrent[j]);
		    double DDy = DDobs[ij] - as_scalar(H * DDmeanCurrent[ij]);
		    DDlogLik(i,j) -= DDy * y + D1y * D2y;
		    if (i != j)
		      DDlogLik(j,i) -= DDy * y + D1y * D2y;
		    ij++;
		  }
	      }
	  }
	if (partialScores)
	  DlogLik.resize(DlogLik.n_rows, DlogLik.n_cols + 1);
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    double Dy = Dobs[i] - as_scalar(H * DmeanCurrent[i]);
	    
	    if (partialScores)
	      DlogLik(i,DlogLik.n_cols-1) = -Dy * y;
	    else
	      DlogLik(i,0) -= Dy * y;
	  }
      }
    if (debugLevel > 5)
      Rcpp::Rcout << "LL+: " << ((S>0.0) ? (- y * y / S - log(S) - LOGTWOPI) / 2.0 : (-y*y)/2.0) << "\ty: " << y << "\tobs[" << series << "," << timeBegin << "]=" << obs << "\tS: " << S << "\tvar: " << var << "\tmean: " << trans(meanCurrent) << endl;

#ifdef USE_LONG
    if (useLong)
      {
        MyMatrix K = covCurrent * H.t() / S;
        MyMatrix HcHC = H * covHC.t();
	MyMatrix HcL = H * covLong;
	meanLong += HcL.t() * (y / S);
	covLong -= K * HcL;

	covLC -= trans(HcHC / S) * HcL;

	covLongDiag -= trans(square(HcL))/S;
      }
#endif

  }


  void addObservation(int series, double timeBegin, double timeEnd, double obs, double var, const MyMatrixRow& exoDataRow, const int seriesIndex, const int houseIndex, const uint k)
  {
    if (timeEnd <= timeBegin)
      {
	addInstantaneousObservation(series, timeBegin, obs, var, exoDataRow, seriesIndex, houseIndex, k);
	return;
      }

    double varPow = 0;
    if ((houseIndex >= 0) & (var > 0))
      varPow = params.getVarCenter(houseIndex, seriesIndex) *
	std::pow(var/params.getVarCenter(houseIndex, seriesIndex),
		 params.getExtraVarPow(houseIndex, seriesIndex));


    vector<double> Dobs;
    vector<double> Dvar;

    for (uint i = 0; i < Dparams.size(); i++)
      {
	Dobs.push_back(0.0);
	Dvar.push_back(0.0);

	Dobs[i] -= Dparams[i].mean(seriesIndex, k < params.mean.n_cols ? k : 0);
	if (houseIndex >= 0)
	  Dobs[i] -= Dparams[i].extraMean(houseIndex, k < params.extraMean.n_cols ? k : 0);

	if (params.hasExo)
	  Dobs[i] -= as_scalar(exoDataRow * Dparams[i].exo);

	if (houseIndex >= 0)
	  {
	    if (var > 0)
	      Dvar[i] = params.getExtraVarMult(houseIndex, seriesIndex)
		* Dparams[i].getExtraVarPow(houseIndex, seriesIndex)
		* log(var/Dparams[i].getVarCenter(houseIndex, seriesIndex))
		* varPow
		+ Dparams[i].getExtraVarMult(houseIndex, seriesIndex)
		* varPow
		+ Dparams[i].getExtraVarAdd(houseIndex, seriesIndex);
	    else
	      Dvar[i] = Dparams[i].getExtraVarAdd(houseIndex, seriesIndex);	      
	  }
      }


    vector<double> DDobs;
    vector<double> DDvar;

    if (hasHess)
      {
	uint ij = 0;
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    for (uint j = i; j < Dparams.size(); j++)
	      {
		DDobs.push_back(0.0);
		DDvar.push_back(0.0);
	    
		DDobs[ij] -= DDparams[ij].mean(seriesIndex, k < params.mean.n_cols ? k : 0);
		if (houseIndex >= 0)
		  DDobs[ij] -= DDparams[ij].extraMean(houseIndex, k < params.extraMean.n_cols ? k : 0);

		if (params.hasExo)
		  DDobs[ij] -= as_scalar(exoDataRow * DDparams[ij].exo);
	    
		if (houseIndex >= 0)
		  {
		    if (var > 0)
		      DDvar[ij] =
			+ Dparams[j].getExtraVarMult(houseIndex, seriesIndex)
			* Dparams[i].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		    
			+ params.getExtraVarMult(houseIndex, seriesIndex)
			* DDparams[ij].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		    
			+ params.getExtraVarMult(houseIndex, seriesIndex)
			* Dparams[i].getExtraVarPow(houseIndex, seriesIndex)
			* Dparams[j].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		
			+ DDparams[ij].getExtraVarMult(houseIndex, seriesIndex)
			* varPow
		    
			+ Dparams[i].getExtraVarMult(houseIndex, seriesIndex)
			* Dparams[j].getExtraVarPow(houseIndex, seriesIndex)
			* log(var/params.getVarCenter(houseIndex, seriesIndex))
			* varPow
		    
			+ DDparams[ij].getExtraVarAdd(houseIndex, seriesIndex);
		    else
		      DDvar[ij] = DDparams[ij].getExtraVarAdd(houseIndex, seriesIndex);

		    ij++;
		  }
	      }
	  }
      }

    obs -= params.mean(seriesIndex, k < params.mean.n_cols ? k : 0);
    if (houseIndex >= 0)
      obs -= params.extraMean(houseIndex, k < params.extraMean.n_cols ? k : 0);

    if (params.hasExo)
      obs -= as_scalar(exoDataRow * params.exo);


    if (houseIndex >= 0)
      {
	var = params.getExtraVarMult(houseIndex, seriesIndex)
	  * varPow
	  + params.getExtraVarAdd(houseIndex, seriesIndex);
      }


    MyMatrix H(1, covHist.n_cols, fill::zeros);
    int j = 0;
    double Hsum = 0.0;

    for (uint i = series; i < covHist.n_cols; i+=seriesSize)
      {
	if (timeIndices[j] > timeEnd)
	  {
	    break;
	  }
	else if (timeIndices[j+1] >= timeBegin)
	  {
	    H(0,i) = max(0.0, min(timeEnd, timeIndices[j+1]) - max(timeBegin, timeIndices[j]));
	    Hsum += H(0,i);
	  }
	j++;
      }
    H /= Hsum;

    double y = obs - as_scalar(H * meanHist);
    double S = as_scalar(H * covHist * H.t()) + var;
    MyMatrix K = covHist * H.t() / S;

    MyMatrix HcHC = H * covHC;

    if (hasHess)
      {
	uint ij = 0;
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    double D1y = Dobs[i] - as_scalar(H * DmeanHist[i]);
	    double D1S = as_scalar(H * DcovHist[i] * H.t()) + Dvar[i];
	    double D1Sinv = - D1S / (S*S);
	    MyMatrix D1K = DcovHist[i] * H.t() / S + covHist * H.t() * D1Sinv;
	    MyMatrix D1HcHC = H * DcovHC[i];
	    for (uint j = i; j < Dparams.size(); j++)
	      {	    
		double D2y = Dobs[j] - as_scalar(H * DmeanHist[j]);
		double D2S = as_scalar(H * DcovHist[j] * H.t()) + Dvar[j];
		double D2Sinv = - D2S / (S*S);
		MyMatrix D2K = DcovHist[j] * H.t() / S + covHist * H.t() * D2Sinv;

		double DDy = DDobs[ij] - as_scalar(H * DDmeanHist[ij]);
		double DDS = as_scalar(H * DDcovHist[ij] * H.t()) + DDvar[ij];
		double DDSinv = 2 * D1S * D2S / (S*S*S) - DDS / (S*S);
		MyMatrix DDK =
		  + DDcovHist[ij] * H.t() / S
		  + DcovHist[i] * H.t() * D2Sinv
		  + DcovHist[j] * H.t() * D1Sinv
		  + covHist * H.t() * DDSinv;
	    
		double DDlogLikChange =
		  + DDy * y / S
		  + D1y * D2y / S
		  + D1y * y * D2Sinv
		  + D2y * y * D1Sinv
		  + (y * y * DDSinv) / 2.0
		  + (DDS / S) / 2.0
		  - (D1S * D2S / (S*S)) / 2.0;
		
		DDlogLik(i,j) -= DDlogLikChange;
		if (i != j)
		  DDlogLik(j,i) -= DDlogLikChange;

		MyMatrix D2HcHC = H * DcovHC[j];
		MyMatrix DDHcHC = H * DDcovHC[ij];

		DDmeanHist[ij] += DDK * y + D1K * D2y + D2K * D1y + K * DDy;
 
		DDmeanCurrent[ij] +=
		  + DDHcHC.t() * (y / S)
		  + D1HcHC.t() * (D2y / S)
		  + D1HcHC.t() * (y * D2Sinv)
	      
		  + D2HcHC.t() * (D1y / S)
		  + HcHC.t() * (DDy / S)
		  + HcHC.t() * (D1y * D2Sinv)
	      
		  + D2HcHC.t() * (y * D1Sinv)
		  + HcHC.t() * (D2y * D1Sinv)
		  + HcHC.t() * (y * DDSinv);

		DDcovHist[ij] -=
		  + DDK * H * covHist
		  + D1K * H * DcovHist[j]
		  + D2K * H * DcovHist[i]
		  + K * H * DDcovHist[ij];
	    
		DDcovCurrent[ij] -=
		  + DDHcHC.t() * HcHC / S
		  + D1HcHC.t() * D2HcHC / S
		  + D1HcHC.t() * HcHC * D2Sinv
	      
		  + D2HcHC.t() * D1HcHC / S
		  + HcHC.t() * DDHcHC / S
		  + HcHC.t() * D1HcHC * D2Sinv
	      
		  + D2HcHC.t() * HcHC * D1Sinv
		  + HcHC.t() * D2HcHC * D1Sinv
		  + HcHC.t() * HcHC * DDSinv;
	    
		DDcovHC[ij] -= DDK * HcHC + D1K * D2HcHC + D2K * D1HcHC + K * DDHcHC;
	    
		ij++;
	      }
	  }
      }

    

    if (partialScores)
      DlogLik.resize(DlogLik.n_rows, DlogLik.n_cols + 1);

    for (uint i = 0; i < Dparams.size(); i++)
      {
	double Dy = Dobs[i] - as_scalar(H * DmeanHist[i]);
	double DS = as_scalar(H * DcovHist[i] * H.t()) + Dvar[i];
	double DSinv = - DS / (S*S);
	MyMatrix DK = DcovHist[i] * H.t() / S + covHist * H.t() * DSinv;

	if (partialScores)
	  DlogLik(i,DlogLik.n_cols-1) = -(Dy * y / S + (y * y * DSinv + DS / S) / 2.0);
	else
	  DlogLik(i,0) -= Dy * y / S + (y * y * DSinv + DS / S) / 2.0;


	MyMatrix DHcHC = H * DcovHC[i];


	DmeanHist[i] += DK * y + K * Dy;
	DmeanCurrent[i] += DHcHC.t() * (y / S) + HcHC.t() * (Dy / S) + HcHC.t() * (y * DSinv);
	DcovHist[i] -= DK * H * covHist + K * H * DcovHist[i];
	DcovCurrent[i] -= DHcHC.t() * HcHC / S + HcHC.t() * DHcHC / S + HcHC.t() * HcHC * DSinv;
	DcovHC[i] -= DK * HcHC + K * DHcHC;

      }

    logLik -= (y * y / S + log(S) + LOGTWOPI) / 2.0;
  

    meanHist += K * y;
    meanCurrent += HcHC.t() * (y / S);
    covHist -= K * H * covHist;
    covCurrent -= HcHC.t() * HcHC / S;
    covHC -= K * HcHC;

#ifdef USE_LONG
    if (useLong)
      {
	MyMatrix HcL = H * covLong;
	meanLong += HcL.t() * (y / S);
	covLong -= K * HcL;

	covLC -= HcHC.t() * HcL / S;

	covLongDiag -= square(HcL.t())/S;
      }
#endif

  }

  MyMatrix getMean(void) const
  {
    return(meanHist);
  }

  MyMatrix getCov(void) const
  {
    return(covHist);
  }

  double getLogLik(void) const
  {
    return(logLik);
  }

  MyMatrix& getDLogLik(void)
  {
    return(DlogLik);
  }

  MyMatrix& getDDLogLik(void)
  {
    return(DDlogLik);
  }

#ifdef USE_LONG
  vector<double> getTimeIndices(void) const
  {
    return(fullTimeIndices);
  }

  MyMatrix getMeanHist(void) const
  {
    assert(useLong);
    return(meanLong);
  }

  MyMatrix getCovHist(void) const
  {
    return(covLongDiag);
  }
#endif
  
  void removeTime(double timeStart)
  {
    uint removeSize = 0;
    vector<double>::iterator removeEnd = timeIndices.begin();
    //while( (*(removeEnd) < timeStart) && (removeEnd+2 != timeIndices.end()) && (removeEnd+1 != timeIndices.end()) && (removeEnd != timeIndices.end()) )
    while( (*(removeEnd) < timeStart) && (removeEnd != timeIndices.end()) )
      {
	removeEnd++;
	removeSize++;
      }
    if (removeSize == 0)
      return;
    removeSize *= seriesSize;

#ifdef USE_LONG
    if (useLong)
      {
	for (vector<double>::iterator ix=timeIndices.begin(); ix != removeEnd; ix++)
	  fullTimeIndices.push_back(*ix);

	meanLong = join_cols(meanLong, meanHist.submat(0, 0, removeSize-1, 0));
	covLong = join_rows(covLong.submat(removeSize, 0, covLong.n_rows-1, covLong.n_cols-1),
			covHist.submat(removeSize, 0, covHist.n_rows-1, removeSize-1));
	covLC = join_rows(covLC, trans(covHC.submat(0, 0, removeSize-1, covHC.n_cols-1)));

	uint cLDr = covLongDiag.n_rows;
	covLongDiag.resize(covLongDiag.n_rows + removeSize, 1);
	for (uint i = 0; i < removeSize; i++)
	  covLongDiag(cLDr + i, 0) = covHist(i, i);
      }
#endif

    timeIndices.erase(timeIndices.begin(), removeEnd);


    for (uint i = 0; i < Dparams.size(); i++)
      {
	DmeanHist[i] = DmeanHist[i].submat(removeSize, 0, DmeanHist[i].n_rows-1, 0);
	DcovHist[i] = DcovHist[i].submat(removeSize, removeSize, DcovHist[i].n_rows-1, DcovHist[i].n_cols-1);
	DcovHC[i] = DcovHC[i].submat(removeSize, 0, DcovHC[i].n_rows-1, DcovHC[i].n_cols-1);
      }

    if (hasHess)
      {
	uint ij = 0;
	for (uint i = 0; i < Dparams.size(); i++)
	  {
	    for (uint j = i; j < Dparams.size(); j++)
	      {
		DDmeanHist[ij] = DDmeanHist[ij].submat(removeSize, 0, DDmeanHist[ij].n_rows-1, 0);
		DDcovHist[ij] = DDcovHist[ij].submat(removeSize, removeSize, DDcovHist[ij].n_rows-1, DDcovHist[ij].n_cols-1);
		DDcovHC[ij] = DDcovHC[ij].submat(removeSize, 0, DDcovHC[ij].n_rows-1, DDcovHC[ij].n_cols-1);
		ij++;
	      }
	  }
      }
    
    meanHist = meanHist.submat(removeSize, 0, meanHist.n_rows-1, 0);
    covHist = covHist.submat(removeSize, removeSize, covHist.n_rows-1, covHist.n_cols-1);
    covHC = covHC.submat(removeSize, 0, covHC.n_rows-1, covHC.n_cols-1);

  }

private:
  bool partialScores;

  double logLik;

  int seriesSize;
  vector<double> timeIndices;

  const OUparam& params;

  const vector<OUparam>& Dparams;

  const vector<OUparam>& DDparams;

  MyMatrix mu;
  MyMatrix P;
  MyMatrix Sigma;

  MyMatrix Ident;
  MyMatrix PpP;
  MyMatrix sPpP;
  MyMatrix sP;

  MatrixExp expP;

  MyMatrix covHist;
  MyMatrix covHC;
  MyMatrix covCurrent;

  MyMatrix meanHist;
  MyMatrix meanCurrent;

  MyMatrix DlogLik;

  MatrixExpFrechet DexpP;

  vector<MyMatrix> DPpP;
  vector<MyMatrix> DsPpP;
  vector<MyMatrix> DsP;
  vector<MyMatrix> Dmu;

  vector<MyMatrix> DmeanHist;
  vector<MyMatrix> DmeanCurrent;
  vector<MyMatrix> DcovHist;
  vector<MyMatrix> DcovHC;
  vector<MyMatrix> DcovCurrent;

  MyMatrix DDlogLik;

  MatrixExpDoubleFrechet DDexpP;

  vector<MyMatrix> DDPpP;
  vector<MyMatrix> DDsPpP;
  vector<MyMatrix> DDsP;
  vector<MyMatrix> DDmu;

  vector<MyMatrix> DDmeanHist;
  vector<MyMatrix> DDmeanCurrent;
  vector<MyMatrix> DDcovHist;
  vector<MyMatrix> DDcovHC;
  vector<MyMatrix> DDcovCurrent;
  
  double lastTime;
  MyMatrix expPnegtime;
  MyMatrix IexpPpPnegtime;

  MyMatrix DexpPnegtime;
  MyMatrix DIexpPpPnegtime;

  MyMatrix DDexpPnegtime;
  MyMatrix DDIexpPpPnegtime;

#ifdef USE_LONG
  bool useLong;
  vector<double> fullTimeIndices;
  MyMatrix meanLong;
  MyMatrix covLong;
  MyMatrix covLC;
  MyMatrix covLongDiag;
#endif

  bool hasHess;
};


class OUsimulation
{
public:
  OUsimulation(const MyMatrix iParams,
	       const inputVector& iTimeIndices,
	       // const inputIntegerVector& iProcessIndices,
	       const int iProcessIndexSize,
	       const int iHouseSize,
	       // const inputIntegerVector& iSeriesIndices,
	       // const inputIntegerVector& iHouseIndices,
	       // const inputVector& iTimeStart,
	       // const inputVector& iTimeEnd,
	       // const MyMatrix& iExoData,
	       // const inputVector& iVariances,
	       // const inputVector& iVarCenters,
	       const bool iHasStart,
	       const bool iHasExo,
	       const bool iVarByHouse,
	       const inputVector& iVarCenters,
	       const inputMatrix& iStartConstraint,
	       const inputMatrix& iMeanConstraint,
	       const inputMatrix& iExtraMeanConstraint,
	       const inputMatrix& iReversionConstraint,
	       const inputMatrix& iSimultaneousConstraint,
	       const inputMatrix& iSigmaSqrtConstraint,
	       const inputMatrix& iExtraVarAddConstraint,
	       const inputMatrix& iExtraVarMultConstraint,
	       const inputMatrix& iExtraVarPowConstraint,
	       const inputMatrix& iExoConstraint)
    :timeIndices(iTimeIndices)
    // ,processIndices(iProcessIndices)
    // ,processIndexSize(iProcessIndexSize)
    // ,seriesIndices(iSeriesIndices)
    // ,houseIndices(iHouseIndices)
    // ,timeStart(iTimeStart)
    // ,timeEnd(iTimeEnd)
    // ,exoData(iExoData)
    // ,variances(iVariances)
    // ,varCenters(iVarCenters)
    ,paramSpace(iProcessIndexSize,
		iHouseSize,
		iHasStart,
		iHasExo,
		iVarByHouse,
		iVarCenters,
		iStartConstraint,
		iMeanConstraint,
		iExtraMeanConstraint,
		iReversionConstraint,
		iSimultaneousConstraint,
		iSigmaSqrtConstraint,
		iExtraVarAddConstraint,
		iExtraVarMultConstraint,
		iExtraVarPowConstraint,
		iExoConstraint)
    ,sampleCurrent(iMeanConstraint.nrow(),timeIndices.size(),iProcessIndexSize,fill::zeros)
    ,sampleHist(iMeanConstraint.nrow(),timeIndices.size()-1,iProcessIndexSize,fill::zeros)
  {
    bool fail;
    params = paramSpace.getParams(iParams, fail);
    vector<OUparam> derivParams;
    vector<OUparam> doubleDerivParams;
    for (uint k = 0; k < ((uint) iProcessIndexSize); k++)
      {
	OUprocess series(params,
			 derivParams,
			 doubleDerivParams,
			 timeIndices(0),
			 false,
			 false);
	sampleCurrent.slice(k).col(0) = series.sampleCurrent();
	if (params.mean.n_cols > 1)
	  sampleCurrent.slice(k).col(0) += params.mean.col(k);
	else
	  sampleCurrent.slice(k).col(0) += params.mean;
	for (int i = 1; i < timeIndices.size(); i++)
	  {
	    series.addTime(timeIndices(i));
	    sampleCurrent.slice(k).col(i) = series.sampleCurrent();
	    if (timeIndices(i)-timeIndices(i-1) > 1e-10)
	      sampleHist.slice(k).col(i-1) = series.sampleHist();
	    else
	      sampleHist.slice(k).col(i-1) = sampleCurrent.slice(k).col(i);
	    if (params.mean.n_cols > 1)
	      {
		sampleCurrent.slice(k).col(i) += params.mean.col(k);
		sampleHist.slice(k).col(i-1) += params.mean.col(k);
	      }
	    else
	      {
		sampleCurrent.slice(k).col(i) += params.mean;
		sampleHist.slice(k).col(i-1) += params.mean;
	      }
	    series.removeTime(timeIndices(i));
	  }
      }
	    
  }

  inputVector getSampleData(const inputIntegerVector& processIndices,
			 const inputIntegerVector& seriesIndices,
			 const inputIntegerVector& houseIndices,
			 const inputVector& timeStart,
			 const inputVector& timeEnd,
			 const MyMatrix& exoData,
			 const inputVector& variances)
  {
    inputVector data(timeStart.size());

    for (int i = 0; i < timeStart.size(); i++)
      {
	double obs = 0;
	double obsLength = 0;
	for (int j = 0; j < timeIndices.size(); j++)
	  {
	    if (timeEnd(i) < timeIndices(j))
	      {
		break;
	      }
	    else
	      {
		if (timeStart(i) == timeIndices(j))
		  {
		    obs = sampleCurrent(seriesIndices(i),j,processIndices(i));
		    break;
		  }
		else if ((timeStart(i) < timeIndices(j)) && (j > 0))
		  {
		    obs += sampleHist(seriesIndices(i),j-1,processIndices(i))
		      * (timeIndices(j) - timeIndices(j-1));
		    obsLength += (timeIndices(j) - timeIndices(j-1));
		  }
	      }
	  }
	if (obsLength > 0)
	  obs /= obsLength;

	if (houseIndices(i) >= 0)
	  obs += params.extraMean(houseIndices(i),
				  (processIndices(i) < ((int) params.extraMean.n_cols)) ? processIndices(i) : 0);
	
	if (params.hasExo)
	  obs += as_scalar(exoData.row(i) * params.exo);

	double var = variances(i);
    	if ((houseIndices(i) >= 0) & (var > 0))
	  {
	    double varPow = params.getVarCenter(houseIndices(i),seriesIndices(i)) *
	      std::pow(var/params.getVarCenter(houseIndices(i),seriesIndices(i)),
		       params.getExtraVarPow(houseIndices(i), seriesIndices(i)));
	    var = params.getExtraVarMult(houseIndices(i), seriesIndices(i))
	      * varPow
	      + params.getExtraVarAdd(houseIndices(i), seriesIndices(i));
	  }
	if (var>0)
	  data(i) = R::rnorm(obs, std::sqrt(var));
	else
	  data(i) = obs;
      }
    return data;
  }

  MyCube& getSampleCurrent()
  {
    return sampleCurrent;
  }
  
  MyCube& getSampleHist()
  {
    return sampleHist;
  }
  
  // MyMatrix getSampleCurrentObs()
  // {
  //   return sampleCurrentObs;
  // }
  
  // MyMatrix getSampleHistObs()
  // {
  //   return sampleHistObs;
  // }
  
private:
  const inputVector& timeIndices;
  // const inputIntegerVector& processIndices;
  // const int processIndexSize;
  // const inputIntegerVector& seriesIndices;
  // const inputIntegerVector& houseIndices;
  // const inputVector& timeStart;
  // const inputVector& timeEnd;
  // const MyMatrix& exoData;
  // const inputVector& variances;
  // const inputVector& varCenters;
  const OUparamspace paramSpace;
  MyCube sampleCurrent;
  MyCube sampleHist;
  // MyMatrix sampleCurrentObs;
  // MyMatrix sampleHistObs;
  OUparam params;
};

inputVector getTimeCumStart(const inputVector& inputTime)
{
  inputVector timeCumStart(inputTime.size());
  if (timeCumStart.size() > 1)
    {
      double nextStart = inputTime(inputTime.size()-1);
      timeCumStart(inputTime.size()-1) = nextStart;
      for (int i = inputTime.size()-2; i >= 0; i--)
	{
	  nextStart = min(inputTime(i), nextStart);
	  timeCumStart(i) = nextStart;
	}
    }
  return timeCumStart;
}

class OUsuper
{
public:
  OUsuper(const inputVector& iTimeIndices,
	  const inputIntegerVector& iProcessIndices,
	  const int iProcessIndexSize,
	  const int iHouseSize,
	  const inputIntegerVector& iSeriesIndices,
	  const inputIntegerVector& iHouseIndices,
	  const inputVector& iTimeStart,
	  const inputVector& iTimeEnd,
	  const inputVector& iTimeCumStart,
	  const MyMatrix& iExoData,
	  const inputVector& iObservations,
	  const inputVector& iVariances,
	  const inputVector& iVarCenters,
	  const bool iHasStart,
	  const bool iHasExo,
	  const bool iVarByHouse,
	  const inputMatrix& iStartConstraint,
	  const inputMatrix& iMeanConstraint,
	  const inputMatrix& iExtraMeanConstraint,
	  const inputMatrix& iReversionConstraint,
	  const inputMatrix& iSimultaneousConstraint,
	  const inputMatrix& iSigmaSqrtConstraint,
	  const inputMatrix& iExtraVarAddConstraint,
	  const inputMatrix& iExtraVarMultConstraint,
	  const inputMatrix& iExtraVarPowConstraint,
	  const inputMatrix& iExoConstraint,
	  const double iEpsilon = 0)
    :mEpsilon(iEpsilon)
    ,timeIndices(iTimeIndices)
    ,processIndices(iProcessIndices)
    ,processIndexSize(iProcessIndexSize)
    ,seriesIndices(iSeriesIndices)
    ,houseIndices(iHouseIndices)
    ,timeStart(iTimeStart)
    ,timeCumStart(iTimeCumStart)
    ,timeEnd(iTimeEnd)
    ,exoData(iExoData)
    ,observations(iObservations)
    ,variances(iVariances)
    ,paramSpace(iProcessIndexSize,
		iHouseSize,
		iHasStart,
		iHasExo,
		iVarByHouse,
		iVarCenters,
		iStartConstraint,
		iMeanConstraint,
		iExtraMeanConstraint,
		iReversionConstraint,
		iSimultaneousConstraint,
		iSigmaSqrtConstraint,
		iExtraVarAddConstraint,
		iExtraVarMultConstraint,
		iExtraVarPowConstraint,
		iExoConstraint)
  {
  }

  OUsuper(const OUsuper& ou)
    :mEpsilon(ou.mEpsilon)
    ,timeIndices(ou.timeIndices)
    ,processIndices(ou.processIndices)
    ,processIndexSize(ou.processIndexSize)
    ,seriesIndices(ou.seriesIndices)
    ,houseIndices(ou.houseIndices)
    ,timeStart(ou.timeStart)
    ,timeCumStart(ou.timeCumStart)
    ,timeEnd(ou.timeEnd)
    ,exoData(ou.exoData)
    ,observations(ou.observations)
    ,variances(ou.variances)
    ,paramSpace(ou.paramSpace)
  {
  }
     
  double getEpsilon(void)
  {
    return mEpsilon;
  }

  void printParams(const MyMatrix& params)
  {
    bool fail;
    OUparam newParams = paramSpace.getParams(params, fail);
    if (fail)
      {
	Rcpp::Rcout << endl << "Failed to set parameters." << endl << endl;
	return;
      }
    Rcpp::Rcout << endl;
    Rcpp::Rcout << (newParams.hasStart ? "Start" : "No start") << endl;
    Rcpp::Rcout << newParams.start.t() << endl;
    Rcpp::Rcout << newParams.mean.t() << endl;
    Rcpp::Rcout << newParams.extraMean.t() << endl;
    Rcpp::Rcout << newParams.reversion << endl;
    Rcpp::Rcout << newParams.sigma << endl;
    Rcpp::Rcout << newParams.extraVarAdd.t() << endl;
    Rcpp::Rcout << newParams.extraVarMult.t() << endl;
    Rcpp::Rcout << newParams.extraVarPow.t() << endl;
    Rcpp::Rcout << (newParams.hasExo ? "Exo" : "No exo") << endl;
    Rcpp::Rcout << newParams.exo.t() << endl;
    Rcpp::Rcout << endl;
  }

  void getHessian(double& logLik, MyMatrix& score,
		MyMatrix& hess,
		const MyMatrix& params, const uint i = 0, const uint j = 0, const bool partialScores = false
		, const bool getHess = false
		) const
  {
    bool fail = false;
    OUparam newParams = paramSpace.getParams(params, fail);
    uint dimScore = 1;
    bool hasScore = j > i;
    if (hasScore)
      {
	dimScore = j - i;
#ifndef MONOCAR_NO_DEBUG
	assert(j <= params.n_rows);
#endif
      }
    if (hasScore)
      {
	if (partialScores)
	  score.set_size(dimScore, 0);
	else
	  score.zeros(dimScore, 1);
	if (getHess)
	  {
	    hess.zeros(dimScore, dimScore);
	  }
      }
    if (fail)
      {
	logLik = -(numeric_limits<double>::infinity());
	return;
      }

    vector<OUparam> derivParams;
    if (hasScore)
      for (uint iDP = i; iDP < j; iDP++)
	derivParams.push_back(paramSpace.getDParams(params, iDP));

    vector<OUparam> doubleDerivParams;
    if (hasScore && getHess)
      for (uint iD1P = i; iD1P < j; iD1P++)
	for (uint iD2P = iD1P; iD2P < j; iD2P++)
	  doubleDerivParams.push_back(paramSpace.getDDParams(params, iD1P, iD2P));

    if (det(newParams.reversion) == 0.0)
      {
	logLik = -(numeric_limits<double>::infinity());
	return;
      }

    try
      {

	logLik = 0.0;

	bool showOutput = false;
	if (partialScores && (debugLevel >= 1) && (debugLevel <= 1))
	  showOutput = true;
	if (getHess && (debugLevel >= 0) && (debugLevel <= 1))
	  showOutput = true;
	long totalIterations = long(processIndexSize) * long(timeStart.size());
	if (totalIterations <= 0)
	  showOutput = false;
	long usedIterations = 0;
	long usedScreen = 0;
	if (showOutput)
	  {
	    Rcpp::Rcout << endl << endl;
	    Rcpp::Rcout << "Computing";
	    if (partialScores && (debugLevel >= 1) && (debugLevel <= 1))
	      {
		Rcpp::Rcout << " partial scores";
		if (getHess)
		  Rcpp::Rcout << " and";
	      }
	    if (getHess && (debugLevel >= 0) && (debugLevel <= 1))
	      Rcpp::Rcout << " Hessian";
	    Rcpp::Rcout << "..." << endl;
	    for (int kx = 0; kx < screenWidth; kx++)
	      Rcpp::Rcout << "_" << flush;
	    Rcpp::Rcout << endl;
	  }

	for (int k = 0; k < processIndexSize; k++)
	  {
	    int timePosition = 0;

	    OUprocess series(newParams,
			     derivParams,
			     doubleDerivParams,
			     timeIndices(timePosition),
			     false,
			     partialScores);

	    for (int i = 0; i < timeStart.size(); i++)
	      {
		if (showOutput)
		  {
		    usedIterations++;
		    while (usedScreen * totalIterations < long(screenWidth) * usedIterations)
		      {
			usedScreen++;
			Rcpp::Rcout << "=" << flush;
		      }
		    R_CheckUserInterrupt();
		  }
		if (timePosition > 0)
		  series.removeTime(min(timeCumStart(i), timeIndices(timePosition-1)));
		while(timeEnd(i) > timeIndices(timePosition) &&
		      timePosition+1 < timeIndices.size())
		  {
		    timePosition++;
		    series.addTime(timeIndices(timePosition));
		  }
		series.removeTime(timeCumStart(i));
		if ((seriesIndices(i) >= 0) && (processIndices(i) == k))
		  {
		    series.addObservation(seriesIndices(i),
					  timeStart(i),
					  timeEnd(i),
					  observations(i),
					  variances(i),
					  exoData.row(i),
					  seriesIndices(i),
					  houseIndices(i),
					  uint(k));

		  }
	      }

	    logLik += series.getLogLik();

	    if (hasScore)
	      {
		if (partialScores)
		  score = join_rows(score, series.getDLogLik());
		else
		  score += series.getDLogLik();
		if (getHess)
		  hess += series.getDDLogLik();
	      }
	  }
		if (showOutput)
		  {
		    usedIterations++;
		    while (usedScreen < long(screenWidth))
		      {
			usedScreen++;
			Rcpp::Rcout << "=" << flush;
		      }
		    Rcpp::Rcout << endl;
		    R_CheckUserInterrupt();
		  }
	return;
      }
    catch (const std::runtime_error& error)
      {
	if (debugLevel > 1)
	  Rcpp::Rcout << "Runtime error in likelihood!" << endl;
	else if (debugLevel >= 0)
	  printDot("!");
	logLik = -(numeric_limits<double>::infinity());
	return;
      }

  }

  void getScore(double& logLik, MyMatrix& score,
		const MyMatrix& params,
		const uint i = 0, const uint j = 0,
		const bool partialScores = false) const
  {
    MyMatrix hess;
    getHessian(logLik, score, hess, params, i, j, partialScores, false);
    return;
  }
  
  double getLikelihood(const MyMatrix& params, const double epsi = 0.0, const uint i = 0, const double epsj = 0.0, const uint j = 0) const
  {
    double logLik;
    MyMatrix score;

    MyMatrix newParams(params);

    if (epsi != 0.0) {
      newParams(i, 0) += epsi;
    }
    if (epsj != 0.0) {
      newParams(j, 0) += epsj;
    }

    getScore(logLik, score, newParams, 0, 0);
    return logLik;
  }
  
#ifdef USE_LONG
  void getHist(const MyMatrix& params,
	       MyMatrix& outputTime,
	       MyMatrix& outputMean,
	       MyMatrix& outputCov)
  {
    bool fail = false;
    OUparam newParams = paramSpace.getParams(params, fail);
    if (fail)
      throw 99;
      
    vector<OUparam> derivParams(0);

    vector<OUparam> doubleDerivParams(0);    

    int timePosition = 0;
    OUprocess series(newParams,
		     derivParams,
		     doubleDerivParams,
		     timeIndices(timePosition),
		     true);

    uint k = 0;
    
    for (int i = 0; i < timeStart.size(); i++)
      {
	if (timePosition > 0)
	  series.removeTime(min(timeCumStart(i), timeIndices(timePosition-1)));
	while(timeEnd(i) > timeIndices(timePosition) &&
	      timePosition+1 < timeIndices.size())
	  {
	    timePosition++;
	    series.addTime(timeIndices(timePosition));
	  }
	series.removeTime(timeCumStart(i));
	if (seriesIndices(i) >= 0)
	  {
	      series.addObservation(seriesIndices(i),
				    timeStart(i),
				    timeEnd(i),
				    observations(i),
				    variances(i),
				    exoData.row(i),
				    seriesIndices(i),
				    houseIndices(i),
				    uint(k));
	  }
      }

    series.removeTime(numeric_limits<double>::infinity());

    MyMatrix outMean =
      reshape(series.getMeanHist(),
	      newParams.reversion.n_rows,
	      series.getMeanHist().n_rows / newParams.reversion.n_rows);
    MyMatrix outCov =
      reshape(series.getCovHist(),
	      newParams.reversion.n_rows,
	      series.getCovHist().n_rows / newParams.reversion.n_rows);
    outCov = sqrt(outCov);

    for (uint i = 0; i < outMean.n_rows; i++) {
      for (uint j = 0; j < outMean.n_cols; j++) {
	outMean(i, j) += newParams.extraMean(i, 0);
      }
    }

    outputTime = vectorToMatrix(series.getTimeIndices());
    outputMean = trans(outMean);
    outputCov = trans(outCov);
  }
#endif

private:
  const double mEpsilon;

  const inputVector timeIndices;
  const inputIntegerVector processIndices;
  const int processIndexSize;
  const inputIntegerVector seriesIndices;
  const inputIntegerVector houseIndices;
  const inputVector timeStart;
  const inputVector timeCumStart;
  const inputVector timeEnd;

  const MyMatrix exoData;

  const inputVector observations;
  const inputVector variances;

  const OUparamspace paramSpace;
};


#ifdef USE_NLOPT
extern "C" {
  double numderiv(OUsuper* ou, MyMatrix& param, double eps, uint i)
  {
    double logLikPlus = -ou->getLikelihood(param, eps, i);
    double logLikMinus = -ou->getLikelihood(param, -eps, i);
    return (logLikPlus - logLikMinus) / (2 * eps);
  }

  double numderivside(OUsuper* ou, MyMatrix& param, double eps, uint i)
  {
    double logLikZero = -ou->getLikelihood(param);
    double logLikPlus = -ou->getLikelihood(param, eps, i);
    return (logLikPlus - logLikZero) / (eps);
  }

  double numsecderiv(OUsuper* ou, MyMatrix param, double eps, uint i, uint j)
  {
    double llpp = -ou->getLikelihood(param, eps, i, eps, j);
    double llpm = -ou->getLikelihood(param, eps, i, -eps, j);
    double llmm = -ou->getLikelihood(param, -eps, i, -eps, j);
    double llmp = -ou->getLikelihood(param, -eps, i, eps, j);
    return ((llpp - llpm)/(eps+eps) + (llmm - llmp)/(eps+eps))/(eps+eps);
  }

  double numsecderivdiag(OUsuper* ou, MyMatrix param, double eps, uint i)
  {
    double llo = -ou->getLikelihood(param);
    double llpp = -ou->getLikelihood(param, eps+eps, i);
    double llmm = -ou->getLikelihood(param, -(eps+eps), i);
    return ((llpp - llo)/(eps+eps) + (llmm - llo)/(eps+eps))/(eps+eps);
  }

  static double oufun(uint n, const double *p, void *ex)
  {
    R_CheckUserInterrupt();

    MyMatrix param(n, 1);
    for (uint i = 0; i < n; i++)
      param(i, 0) = p[i];
    
    OUsuper* ou = (OUsuper*) ex;
    double logLik = ou->getLikelihood(param);
    if (debugLevel > 4)
      ou->printParams(param);
    if (debugLevel > 3)
      Rcpp::Rcout << "Log-likelihood: " << logLik << endl;
    return(-logLik);
  }

  static void ougrad_analytic(uint n, const double *p, double *gr, void *ex)
  {
    MyMatrix param(n, 1);
    for (uint i = 0; i < n; i++)
      param(i, 0) = p[i];
    
    OUsuper* ou = (OUsuper*) ex;

    if (debugLevel > 2)
      ou->printParams(param);

    double logLik;
    MyMatrix score;

    ou->getScore(logLik, score, param, 0, n);    
#ifndef MONOCAR_NO_DEBUG
    assert(n == score.n_rows);
#endif
    for (uint i = 0; i < n; i++)
      gr[i] = -score(i, 0);
  }

  static void ougrad_numeric(uint n, const double *p, double *gr, void *ex)
  {
    MyMatrix param(n, 1);
    for (uint i = 0; i < n; i++)
      param(i, 0) = p[i];
    
    OUsuper* ou = (OUsuper*) ex;

    const double eps = ou->getEpsilon();

    if (debugLevel > 2)
      ou->printParams(param);
      
    if (debugLevel > 1)
      Rcpp::Rcout << "Grad (n):" << flush;

    R_CheckUserInterrupt();
    for (uint i = 0; i < n; i++)
      {
	double grvalue = (4 * numderiv(ou, param, eps/2, i) - numderiv(ou, param, eps, i))/3;
	if (!isfinite(grvalue))
	  {
	    if (debugLevel > 1)
	      Rcpp::Rcout << "nan->" << flush;
	    grvalue = 2 * numderivside(ou, param, eps/2, i) - numderivside(ou, param, eps, i);
	    if (!isfinite(grvalue))
	      {
		if (debugLevel > 1)
		  Rcpp::Rcout << "nan->" << flush;
		grvalue = 2 * numderivside(ou, param, -eps/2, i) - numderivside(ou, param, -eps, i);
	      }
	  }
	if (debugLevel > 1)
	  Rcpp::Rcout << " " << grvalue << flush;

	gr[i] = grvalue;
      }

    if (debugLevel > 1)
      Rcpp::Rcout << endl;
  }

  static void ougrad(uint n, const double *p, double *gr, void *ex)
    {
      OUsuper* ou = (OUsuper*) ex;
      const double eps = ou->getEpsilon();

      if (eps <= 0.0)
	ougrad_analytic(n, p, gr, ex);
      else
	ougrad_numeric(n, p, gr, ex);

    if (debugLevel > 1)
      {
	Rcpp::Rcout << "Gradient:" << flush;
	for (uint i = 0; i < n; i++)
	  Rcpp::Rcout << " " << gr[i] << flush;
	Rcpp::Rcout << endl;
      }
    }
  
  static void ougradloud(uint n, const double *p, double *gr, void *ex)
    {
      if (debugLevel == 1)
	printDot(":");
      if (debugLevel == 0)
	printDot(".");
      ougrad(n, p, gr, ex);
    }

  static void ouhess_analytic(uint n, const double *p,
		     double *hess,
		     void *ex)
  {
    MyMatrix param(n, 1);
    for (uint i = 0; i < n; i++)
      param(i, 0) = p[i];
    
    OUsuper* ou = (OUsuper*) ex;

    if (debugLevel > 2)
      ou->printParams(param);

    if (debugLevel > 1) {
      Rcpp::Rcout << "Computing Hessian" << endl;
      // for (uint i = 0; i < n; i++)
      // 	for (uint j = i; j < n; j++)
      // 	  Rcpp::Rcout << "_";
      // Rcpp::Rcout << endl;
    }

    double logLik;
    MyMatrix score;
    MyMatrix hessianMatrix;

    ou->getHessian(logLik, score, hessianMatrix, param, 0, n, false, true);
#ifndef MONOCAR_NO_DEBUG
    assert(n == score.n_rows);
#endif
    // for (uint i = 0; i < n; i++)
    //   gr[i] = -score(i, 0);
    uint ij = 0;
    for (uint i = 0; i < n; i++)
      {
	for (uint j = 0; j < n; j++)
	  {
	    hess[ij] = -hessianMatrix(i, j);
	    ij++;
	  }
      }

    R_CheckUserInterrupt();
    if (debugLevel > 1)
      printDot(".");
    if (debugLevel == 1 || debugLevel == 0)
      printDot("*");
  }

  static void ouhess_numeric(uint n, const double *p,
		     double *hess,
		     void *ex)
  {
    OUsuper* ou = (OUsuper*) ex;

    const double eps = ou->getEpsilon();

#ifndef MONOCAR_NO_DEBUG
      assert(eps != 0);
#endif

    if (debugLevel > 1) {
      Rcpp::Rcout << "Computing Hessian" << endl;
      for (uint i = 0; i < n; i++)
    	for (uint j = i; j < n; j++)
    	  Rcpp::Rcout << "_";
      Rcpp::Rcout << endl;
    }

    MyMatrix param(n, 1);
    for (uint i = 0; i < n; i++)
      param(i, 0) = p[i];


      for (uint i = 0; i < n; i++)
	{
	  hess[i+i*n] = (4 * numsecderivdiag(ou, param, eps/2, i) - numsecderivdiag(ou, param, eps, i))/3;
	  for (uint j = i+1; j < n; j++)
	    {
	      R_CheckUserInterrupt();
	      hess[i+j*n] = (4 * numsecderiv(ou, param, eps/2, i, j) - numsecderiv(ou, param, eps, i, j))/3;
	      hess[j+i*n] = hess[i+j*n];
	    }
	  if (debugLevel > 1)
	    printDot(".");
	  if (debugLevel == 1 || debugLevel == 0)
	    printDot("*");
	}
      
      if (debugLevel > 1)
	Rcpp::Rcout << endl;
  }

  static void ouhess(uint n, const double *p,
		     double *hess,
		     void *ex)
    {
      OUsuper* ou = (OUsuper*) ex;
      const double eps = ou->getEpsilon();
      if (eps <= 0.0)
	ouhess_analytic(n, p, hess, ex);
      else
	ouhess_numeric(n, p, hess, ex);
    }

}

static double oufg(unsigned n, const double *xin, double *grad, void *ex)
{
  double loglik = oufun(n, xin, ex);
  if (grad) {
    ougrad(n, xin, grad, ex);
    if (debugLevel == 1)
      printDot(":");
    if (debugLevel == 0)
      printDot(".");
  }
  else {
    if (debugLevel == 1)
      printDot(".");
  }
  return loglik;
}
#endif


RcppExport SEXP monocarptr__new(SEXP iParam, SEXP iOutput,
			   SEXP iOptimizers, SEXP iMaxIters,
			   SEXP iTolerances, SEXP iEpsilon,
			   SEXP iLowerBounds, SEXP iUpperBounds,
			   SEXP iStart, SEXP iMean, SEXP iExtraMean,
			   SEXP iReversion, SEXP iSimultaneous, SEXP iSigma,
			   SEXP iExtraVarAdd, SEXP iExtraVarMult, SEXP iExtraVarPow,
			   SEXP iExo,
			   SEXP iTimeIndices, SEXP iProcessIndices,
			   SEXP iSeriesIndices, SEXP iHouseIndices,
			   SEXP iTimeStart, SEXP iTimeEnd,
			   SEXP iExoData,
			   SEXP iObservations, SEXP iVariances,
			   SEXP iVarCenters, SEXP iSizes, SEXP iScreenWidth)
{
#ifdef TRY_X
  char *exceptionMesg=NULL;

  try {
#endif
    
    GetRNGstate();

    inputVector param(iParam);

    inputVector sizes(iSizes);
#ifndef MONOCAR_NO_DEBUG
    assert(sizes.size()>=4);
#endif
    int processIndexSize = sizes(0);
    //int seriesIndexSize = sizes(1);
    int houseIndexSize = sizes(2);
    bool varByHouse = sizes(3) > 0;
    
    inputIntegerVector outputInfo(iOutput);
    vector<string> optimizers = Rcpp::as<vector<string> >(iOptimizers);
    vector<int> maxiters = Rcpp::as<vector<int> >(iMaxIters);
    vector<double> tolerances = Rcpp::as<vector<double> >(iTolerances);
    double epsilon = Rcpp::as<double>(iEpsilon);

    vector<double> lowerBounds(0);
    if (!Rf_isNull(iLowerBounds))
      lowerBounds = Rcpp::as<vector<double> >(iLowerBounds);

    vector<double> upperBounds(0);
    if (!Rf_isNull(iUpperBounds))
      upperBounds = Rcpp::as<vector<double> >(iUpperBounds);

    bool hasStart = !Rf_isNull(iStart);
    inputMatrix start;
    if (hasStart)
      start = inputMatrix(iStart);
    else
      start = inputMatrix(iMean);
    inputMatrix mean(iMean);
    inputMatrix extraMean(iExtraMean);
    inputMatrix reversion(iReversion);
    inputMatrix simultaneous(iSimultaneous);
    inputMatrix sigma(iSigma);
    inputMatrix extraVarAdd(iExtraVarAdd);
    inputMatrix extraVarMult(iExtraVarMult);
    inputMatrix extraVarPow(iExtraVarPow);
    inputMatrix exo(0,reversion.ncol());
    bool hasExo = !Rf_isNull(iExo);
    if (hasExo)
      exo = inputMatrix(iExo);
    
    inputVector timeIndices(iTimeIndices);    
    inputIntegerVector processIndices(iProcessIndices);
    inputIntegerVector seriesIndices(iSeriesIndices);
    inputIntegerVector houseIndices(iHouseIndices);
    inputVector timeStart(iTimeStart);
    inputVector timeEnd(iTimeEnd);
    inputVector observations(iObservations);
    inputVector variances(iVariances);
    inputVector varCenters(iVarCenters);

    MyMatrix exoData;
    if (hasExo)
      {
	if (!Rf_isNull(iExoData))
	  {
	    exoData = Rcpp::as<MyMatrix>(iExoData);
	    assert(exoData.n_cols == uint(exo.nrow()));
	  }
	else
	  hasExo = false;
      }
    
    MyMatrix paramMat = RtoMatrix(param);

#ifndef MONOCAR_NO_DEBUG
    assert(outputInfo.size() >= 3);
#endif

    if (outputInfo.size() >= 4)
      debugLevel = outputInfo(3) - 1;
    else
      debugLevel = 0;

    screenWidth = Rcpp::as<int>(iScreenWidth);
    screenUse = 0;

    //int processIndexSize = maxIndex(processIndices)+1;

    inputVector timeCumStart = getTimeCumStart(timeStart);
    Rcpp::XPtr<OUsuper>
      ptr(new OUsuper(timeIndices,
		      processIndices,
		      processIndexSize,
		      houseIndexSize,
		      seriesIndices,
		      houseIndices,
		      timeStart,
		      timeEnd,
		      timeCumStart,
		      exoData,
		      observations,
		      variances,
		      varCenters,
		      hasStart,
		      hasExo,
		      varByHouse,
		      start,
		      mean,
		      extraMean,
		      reversion,
		      simultaneous,
		      sigma,
		      extraVarAdd,
		      extraVarMult,
		      extraVarPow,
		      exo,
		      epsilon),
	  true );
    
    // return the external pointer to the R side
    return ptr;
    
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }
    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}

RcppExport SEXP monocarptr__fun( SEXP xp, SEXP param_ )
{
#ifdef TRY_X
  char *exceptionMesg=NULL;
  try {
#endif

    Rcpp::XPtr<OUsuper> ou(xp);
    //MyMatrixCol param = Rcpp::as<MyMatrixCol>(param_);
    inputVector param(param_);
    MyMatrix paramMat = RtoMatrix(param);
    
    double negLogLik = - ou->getLikelihood(paramMat);

    if (debugLevel > 4)
      ou->printParams(param);
    
    return Rcpp::wrap(negLogLik);
      
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}


double numderiv(Rcpp::XPtr<OUsuper> ou, MyMatrix& param, double eps, uint i)
{
  double logLikPlus = -ou->getLikelihood(param, eps, i);
  double logLikMinus = -ou->getLikelihood(param, -eps, i);
  return (logLikPlus - logLikMinus) / (2 * eps);
}

double numderivside(Rcpp::XPtr<OUsuper> ou, MyMatrix& param, double eps, uint i)
{
  double logLikZero = -ou->getLikelihood(param);
  double logLikPlus = -ou->getLikelihood(param, eps, i);
  return (logLikPlus - logLikZero) / (eps);
}

double numsecderiv(Rcpp::XPtr<OUsuper> ou, MyMatrix& param, double eps, uint i, uint j)
{
  double llpp = -ou->getLikelihood(param, eps, i, eps, j);
  double llpm = -ou->getLikelihood(param, eps, i, -eps, j);
  double llmm = -ou->getLikelihood(param, -eps, i, -eps, j);
  double llmp = -ou->getLikelihood(param, -eps, i, eps, j);
  return ((llpp - llpm)/(eps+eps) + (llmm - llmp)/(eps+eps))/(eps+eps);
}

double numsecderivdiag(Rcpp::XPtr<OUsuper> ou, MyMatrix& param, double eps, uint i)
{
  double llo = -ou->getLikelihood(param);
  double llpp = -ou->getLikelihood(param, eps+eps, i);
  double llmm = -ou->getLikelihood(param, -(eps+eps), i);
  return ((llpp - llo)/(eps+eps) + (llmm - llo)/(eps+eps))/(eps+eps);
}


RcppExport SEXP monocarptr__grad( SEXP xp, SEXP param_ )
{
#ifdef TRY_X
  char *exceptionMesg=NULL;
  try {
#endif

    Rcpp::XPtr<OUsuper> ou(xp);
    // MyMatrixCol param = Rcpp::as<MyMatrixCol>(param_);
    inputVector paramX(param_);
    MyMatrix param = RtoMatrix(paramX);
    // double *x = param.memptr();
    double eps = ou->getEpsilon();
    uint n = param.n_rows;
    
    double logLik;
    MyMatrixCol gr(n);

    if (debugLevel == 1)
      printDot(":");
    if (debugLevel == 0)
      printDot(".");
    
    if (eps <= 0.0) {      
    MyMatrix score;
    ou->getScore(logLik, score, param, 0, n);    
#ifndef MONOCAR_NO_DEBUG
    assert(n == score.n_rows);
#endif
    for (uint i = 0; i < n; i++)
      gr[i] = -score(i, 0);
    }
    else
      {
    for (uint i = 0; i < n; i++)
      {
	double grvalue = (4 * numderiv(ou, param, eps/2, i) - numderiv(ou, param, eps, i))/3;
	if (!isfinite(grvalue))
	  {
	    if (debugLevel > 1)
	      Rcpp::Rcout << "nan->" << flush;
	    grvalue = 2 * numderivside(ou, param, eps/2, i) - numderivside(ou, param, eps, i);
	    if (!isfinite(grvalue))
	      {
		if (debugLevel > 1)
		  Rcpp::Rcout << "nan->" << flush;
		grvalue = 2 * numderivside(ou, param, -eps/2, i) - numderivside(ou, param, -eps, i);
	      }
	  }
	if (debugLevel > 1)
	  Rcpp::Rcout << " " << grvalue << flush;

	gr[i] = grvalue;
      }
      }
    
    return Rcpp::wrap(gr);
      
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}







RcppExport SEXP monocarptr__partial( SEXP xp, SEXP param_ )
{
#ifdef TRY_X
  char *exceptionMesg=NULL;
  try {
#endif

    Rcpp::XPtr<OUsuper> ou(xp);
    // MyMatrixCol param = Rcpp::as<MyMatrixCol>(param_);
    inputVector param(param_);
    MyMatrix paramMat = RtoMatrix(param);
    double outLogLik;
    MyMatrix outPartialScores;
    
    ou->getScore(outLogLik, outPartialScores, paramMat, 0, paramMat.n_rows, true);

    Rcpp::List ret;
    ret["grad"] = outPartialScores.t();
    ret["logLik"] = outLogLik;
    return ret;
      
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}


RcppExport SEXP monocarptr__hess( SEXP xp, SEXP param_ )
{
#ifdef TRY_X
  char *exceptionMesg=NULL;
  try {
#endif

    Rcpp::XPtr<OUsuper> ou(xp);
    // MyMatrixCol param = Rcpp::as<MyMatrixCol>(param_);
    inputVector paramX(param_);
    MyMatrix param = RtoMatrix(paramX);
    uint n = param.n_rows;
    double eps = ou->getEpsilon();

    MyMatrix hessianMatrix(n,n);

    if (debugLevel > 2)
      ou->printParams(param);

    if (eps <= 0.0)
      {

	if (debugLevel > 1) {
	  Rcpp::Rcout << "Computing Hessian" << endl;
	}

	double logLik;
	MyMatrix score;

	ou->getHessian(logLik, score, hessianMatrix, param, 0, n, false, true);
#ifndef MONOCAR_NO_DEBUG
	assert(n == score.n_rows);
#endif

	R_CheckUserInterrupt();
	// if (debugLevel > 1)
	//   printDot(".");
	// if (debugLevel == 1 || debugLevel == 0)
	//   printDot("*");
	
      }
    else
      {
#ifndef MONOCAR_NO_DEBUG
      assert(eps != 0);
#endif

    if (debugLevel > 1) {
      Rcpp::Rcout << "Computing Hessian" << endl;
      for (uint i = 0; i < n; i++)
    	for (uint j = i; j < n; j++)
    	  Rcpp::Rcout << "_";
      Rcpp::Rcout << endl;
    }


      for (uint i = 0; i < n; i++)
	{
	  hessianMatrix(i,i) = (4 * numsecderivdiag(ou, param, eps/2, i) - numsecderivdiag(ou, param, eps, i))/3;
	  for (uint j = i+1; j < n; j++)
	    {
	      R_CheckUserInterrupt();
	      hessianMatrix(i,j) = (4 * numsecderiv(ou, param, eps/2, i, j) - numsecderiv(ou, param, eps, i, j))/3;
	      hessianMatrix(j,i) = hessianMatrix(i,j);
	    }
	  if (debugLevel > 1)
	    printDot(".");
	  if (debugLevel == 1 || debugLevel == 0)
	    printDot("*");
	}
      
      if (debugLevel > 1)
	Rcpp::Rcout << endl;
      }

    hessianMatrix *= -1.0;

    return Rcpp::wrap(hessianMatrix);
      
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}


RcppExport SEXP monocarptr__hist( SEXP xp, SEXP param_ )
{
#ifdef TRY_X
  char *exceptionMesg=NULL;
  try {
#endif

    Rcpp::XPtr<OUsuper> ou(xp);
    inputVector param(param_);
    MyMatrix paramMat = RtoMatrix(param);

    MyMatrix timeMat;
    MyMatrix meanMat;
    MyMatrix covMat;
    ou->getHist(paramMat, timeMat, meanMat, covMat);
	
    Rcpp::List ret;
    ret["hist.time"] = timeMat;
    ret["hist.mean"] = meanMat;
    ret["hist.cov"] = covMat;
    
    return ret;
      
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}

RcppExport SEXP monocarptr__endopt( SEXP success_ )
{
  int success = Rcpp::as<int>(success_);
  if (success<=0)
    {
      if ((debugLevel == 1) | (debugLevel == 0))
	printDot("-");
      else if (debugLevel > 1)
	Rcpp::Rcout << "nlopt failed!" << endl;
    }
  else
    {
      if ((debugLevel == 1) | (debugLevel == 0))
	printDot("+");
      else if (debugLevel > 1)
	Rcpp::Rcout << "found minimum" << endl;
    }
    Rcpp::List ret;
    return ret;
}

#ifdef USE_NLOPT
RcppExport SEXP monocarest(SEXP iParam, SEXP iOutput,
			   SEXP iOptimizers, SEXP iMaxIters,
			   SEXP iTolerances, SEXP iEpsilon,
			   SEXP iLowerBounds, SEXP iUpperBounds,
			   SEXP iStart, SEXP iMean, SEXP iExtraMean,
			   SEXP iReversion, SEXP iSimultaneous, SEXP iSigma,
			   SEXP iExtraVarAdd, SEXP iExtraVarMult, SEXP iExtraVarPow,
			   SEXP iExo,
			   SEXP iTimeIndices, SEXP iProcessIndices,
			   SEXP iSeriesIndices, SEXP iHouseIndices,
			   SEXP iTimeStart, SEXP iTimeEnd,
			   SEXP iExoData,
			   SEXP iObservations, SEXP iVariances,
			   SEXP iVarCenters, SEXP iSizes, SEXP iScreenWidth)
{
#ifdef TRY_X
  char *exceptionMesg=NULL;

  try {
#endif
    
    GetRNGstate();
    nlopt_srand((unsigned long) unif_rand()*numeric_limits<unsigned long>::max());

    inputVector param(iParam);

    inputVector sizes(iSizes);
#ifndef MONOCAR_NO_DEBUG
    assert(sizes.size()>=4);
#endif
    int processIndexSize = sizes(0);
    //int seriesIndexSize = sizes(1);
    int houseIndexSize = sizes(2);
    bool varByHouse = sizes(3) > 0;
    
    inputIntegerVector outputInfo(iOutput);
    vector<string> optimizers = Rcpp::as<vector<string> >(iOptimizers);
    vector<int> maxiters = Rcpp::as<vector<int> >(iMaxIters);
    vector<double> tolerances = Rcpp::as<vector<double> >(iTolerances);
    double epsilon = Rcpp::as<double>(iEpsilon);

    vector<double> lowerBounds(0);
    if (!Rf_isNull(iLowerBounds))
      lowerBounds = Rcpp::as<vector<double> >(iLowerBounds);

    vector<double> upperBounds(0);
    if (!Rf_isNull(iUpperBounds))
      upperBounds = Rcpp::as<vector<double> >(iUpperBounds);

    bool hasStart = !Rf_isNull(iStart);
    inputMatrix start;
    if (hasStart)
      start = inputMatrix(iStart);
    else
      start = inputMatrix(iMean);
    inputMatrix mean(iMean);
    inputMatrix extraMean(iExtraMean);
    inputMatrix reversion(iReversion);
    inputMatrix simultaneous(iSimultaneous);
    inputMatrix sigma(iSigma);
    inputMatrix extraVarAdd(iExtraVarAdd);
    inputMatrix extraVarMult(iExtraVarMult);
    inputMatrix extraVarPow(iExtraVarPow);
    inputMatrix exo(0,reversion.ncol());
    bool hasExo = !Rf_isNull(iExo);
    if (hasExo)
      exo = inputMatrix(iExo);
    
    inputVector timeIndices(iTimeIndices);    
    inputIntegerVector processIndices(iProcessIndices);
    inputIntegerVector seriesIndices(iSeriesIndices);
    inputIntegerVector houseIndices(iHouseIndices);
    inputVector timeStart(iTimeStart);
    inputVector timeEnd(iTimeEnd);
    inputVector observations(iObservations);
    inputVector variances(iVariances);
    inputVector varCenters(iVarCenters);

    MyMatrix exoData;
    if (hasExo)
      {
	if (!Rf_isNull(iExoData))
	  {
	    exoData = Rcpp::as<MyMatrix>(iExoData);
	    assert(exoData.n_cols == uint(exo.nrow()));
	  }
	else
	  hasExo = false;
      }
    
    MyMatrix paramMat = RtoMatrix(param);

#ifndef MONOCAR_NO_DEBUG
    assert(outputInfo.size() >= 3);
#endif

    if (outputInfo.size() >= 4)
      debugLevel = outputInfo(3) - 1;
    else
      debugLevel = 0;

    screenWidth = Rcpp::as<int>(iScreenWidth);
    screenUse = 0;

    //int processIndexSize = maxIndex(processIndices)+1;

    inputVector timeCumStart = getTimeCumStart(timeStart);
    OUsuper ou(timeIndices,
	       processIndices,
	       processIndexSize,
	       houseIndexSize,
	       seriesIndices,
	       houseIndices,
	       timeStart,
	       timeEnd,
	       timeCumStart,
	       exoData,
	       observations,
	       variances,
	       varCenters,
	       hasStart,
	       hasExo,
	       varByHouse,
	       start,
	       mean,
	       extraMean,
	       reversion,
	       simultaneous,
	       sigma,
	       extraVarAdd,
	       extraVarMult,
	       extraVarPow,
	       exo,
	       epsilon
	       );


    double *x = paramMat.memptr();
    double min = 1;

    bool getGrad = false;
    bool getHess = false;
    bool getHist = false;
    if (outputInfo.size() > 0)
      getGrad = outputInfo(0) > 0;
    if (outputInfo.size() > 1)
      getHess = outputInfo(1) > 0;
    if (outputInfo.size() > 2)
      getHist = outputInfo(2) > 0;

    
    for (uint i = 0; i < optimizers.size(); i++)
      {
	bool hasOptimizer = false;
	nlopt_opt opt;
	string optimizerName = optimizers[i];
	if (optimizerName == "DIRECT" ||
	    optimizerName == "DIRECTL" ||
	    optimizerName == "DIRECT_L")
	  opt = nlopt_create(NLOPT_GN_DIRECT_L, paramMat.n_rows);
	if (optimizerName == "CRS" || optimizerName == "CRS2")
	  {
	    opt = nlopt_create(NLOPT_GN_CRS2_LM, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "MLSL")
	  {
	    opt = nlopt_create(NLOPT_G_MLSL_LDS, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "STOGO")
	  {
	    opt = nlopt_create(NLOPT_GD_STOGO, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "STOGORAND" || optimizerName == "STOGO_RAND")
	  {
	    opt = nlopt_create(NLOPT_GD_STOGO_RAND, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "ISRES")
	  {
	    opt = nlopt_create(NLOPT_GN_ISRES, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "ESCH")
	  {
	    opt = nlopt_create(NLOPT_GN_ESCH, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "COBYLA")
	  {
	    opt = nlopt_create(NLOPT_LN_COBYLA, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "BOBYQA")
	  {
	    opt = nlopt_create(NLOPT_LN_BOBYQA, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "NEWUOA")
	  {
	    opt = nlopt_create(NLOPT_LN_NEWUOA, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "PRAXIS")
	  {
	    opt = nlopt_create(NLOPT_LN_PRAXIS, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "SUBPLEX" || optimizerName == "SBPLX")
	  {
	    opt = nlopt_create(NLOPT_LN_SBPLX, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "NM" || optimizerName == "NELDERMEAD")
	  {
	    opt = nlopt_create(NLOPT_LN_NELDERMEAD, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "MMA")
	  {
	    opt = nlopt_create(NLOPT_LD_MMA, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "SLSQP")
	  {
	    opt = nlopt_create(NLOPT_LD_SLSQP, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "BFGS" || optimizerName == "LBFGS")
	  {
	    opt = nlopt_create(NLOPT_LD_LBFGS, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "TNEWTON")
	  {
	    opt = nlopt_create(NLOPT_LD_TNEWTON_RESTART, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "VAR2")
	  {
	    opt = nlopt_create(NLOPT_LD_VAR2, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	if (optimizerName == "VAR1")
	  {
	    opt = nlopt_create(NLOPT_LD_VAR1, paramMat.n_rows);
	    hasOptimizer = true;
	  }
	    
	if (!hasOptimizer)
	  {
	    Rcpp::Rcout<< "Skipping unknown optimizer: " << optimizerName << endl;
	  }
	else {
	  nlopt_set_min_objective(opt, oufg, &ou);

	  if (lowerBounds.size() == paramMat.n_rows)
	    {
	      nlopt_set_lower_bounds(opt, lowerBounds.data());
	    }
	  else if (lowerBounds.size() > 0)
	    {
	      Rcpp::Rcout << "Invalid lower bounds: bound of size " <<
		lowerBounds.size() << " but should be size " <<
		paramMat.n_rows << endl;
	    }

	  if (upperBounds.size() == paramMat.n_rows)
	    {
	      nlopt_set_upper_bounds(opt, upperBounds.data());
	    }
	  else if (upperBounds.size() > 0)
	    {
	      Rcpp::Rcout << "Invalid upper bounds: bound of size " <<
		upperBounds.size() << " but should be size " <<
		paramMat.n_rows << endl;
	    }

	  
	  double tol = 1e-3;
	  if (i < tolerances.size())
	    tol = tolerances[i];
	  nlopt_set_xtol_rel(opt, tol);
	  
	  if (i < maxiters.size())
	    if (maxiters[i] >= 0)
	      nlopt_set_maxeval(opt, maxiters[i]);
	  
	  if (nlopt_optimize(opt, x, &min) < 0)
	    {
	      if ((debugLevel == 1) | (debugLevel == 0))
		printDot("-");
	      else if (debugLevel > 1)
		Rcpp::Rcout << "nlopt failed!" << endl;
	    }
	  else
	    {
	      if ((debugLevel == 1) | (debugLevel == 0))
		printDot("+");
	      else if (debugLevel > 1)
		Rcpp::Rcout << "found minimum" << endl;
	    }
	  
	  nlopt_destroy(opt);
	  }
      }
     
    Rcpp::List ret;

    ret["min"] = min;
    //inputMatrix outParam=MatrixToR(paramMat);
    ret["par"] = paramMat;

    if (getGrad)
      {
	double outLogLik;
	MyMatrix outPartialScores;

	ou.getScore(outLogLik, outPartialScores, paramMat, 0, paramMat.n_rows, true);

	MyMatrix scoreMat(paramMat.n_rows, 1);
	double *scorePtr = scoreMat.memptr();
	ougradloud(paramMat.n_rows, x, scorePtr, &ou);
	//inputMatrix outScore=MatrixToR(scoreMat);
	ret["score"] = scoreMat;
	
	//inputMatrix outGrad=MatrixToR(outPartialScores.t());
	ret["grad"] = outPartialScores.t();
	ret["logLik"] = outLogLik;
      }
    else
      {
	ret["logLik"] = ou.getLikelihood(paramMat);
      }

    if (getHess)
      {
	MyMatrix hessMat(paramMat.n_rows, paramMat.n_rows);
	double *hess = hessMat.memptr();
	ouhess(paramMat.n_rows, x, hess, &ou);

	//inputMatrix outHess=MatrixToR(hessMat);
	ret["hess"] = hessMat;
      }

    if(getHist)
      {
	MyMatrix timeMat;
	MyMatrix meanMat;
	MyMatrix covMat;
	ou.getHist(paramMat, timeMat, meanMat, covMat);
	// inputMatrix outTime=MatrixToR(timeMat);
	// inputMatrix outMean=MatrixToR(meanMat);
	// inputMatrix outCov=MatrixToR(covMat);
	ret["hist.time"] = timeMat;
	ret["hist.mean"] = meanMat;
	ret["hist.cov"] = covMat;
      }


    Rcpp::Rcout << endl;
    PutRNGstate();
    return ret;
	
#ifdef TRY_X
  } catch(std::exception& ex) {
    exceptionMesg = copyMessageToR(ex.what());
  } catch(...) {
    exceptionMesg = copyMessageToR("unknown reason");
  }
    
  if(exceptionMesg != NULL)
    Rf_error(exceptionMesg);
#endif

}
#endif


RcppExport SEXP ousim(SEXP iParam,
		      SEXP iStart, SEXP iMean, SEXP iExtraMean,
		      SEXP iReversion, SEXP iSimultaneous, SEXP iSigma,
		      SEXP iExtraVarAdd, SEXP iExtraVarMult, SEXP iExtraVarPow,
		      SEXP iExo,
		      SEXP iTimeIndices, SEXP iProcessIndices,
		      SEXP iSeriesIndices, SEXP iHouseIndices,
		      SEXP iTimeStart, SEXP iTimeEnd,
		      SEXP iExoData,
		      SEXP iVariances, SEXP iVarCenters,
		      SEXP iSizes, SEXP iScreenWidth)
{    
    GetRNGstate();
    inputVector param(iParam);
    
    debugLevel = 0;

    screenWidth = Rcpp::as<int>(iScreenWidth);
    screenUse = 0;


    inputVector sizes(iSizes);
#ifndef MONOCAR_NO_DEBUG
    assert(sizes.size()>=4);
#endif
    int processIndexSize = sizes(0);
    //int seriesIndexSize = sizes(1);
    int houseIndexSize = sizes(2);
    bool varByHouse = sizes(3) > 0;

        bool hasStart = !Rf_isNull(iStart);
    inputMatrix start;
    if (hasStart)
      start = inputMatrix(iStart);
    else
      start = inputMatrix(iMean);
    inputMatrix mean(iMean);
    inputMatrix extraMean(iExtraMean);
    inputMatrix reversion(iReversion);
    inputMatrix simultaneous(iSimultaneous);
    inputMatrix sigma(iSigma);
    inputMatrix extraVarAdd(iExtraVarAdd);
    inputMatrix extraVarMult(iExtraVarMult);
    inputMatrix extraVarPow(iExtraVarPow);
    inputMatrix exo(0,reversion.ncol());
    bool hasExo = !Rf_isNull(iExo);
    if (hasExo)
      exo = inputMatrix(iExo);
    
    inputVector timeIndices(iTimeIndices);    
    inputIntegerVector processIndices(iProcessIndices);
    inputIntegerVector seriesIndices(iSeriesIndices);
    inputIntegerVector houseIndices(iHouseIndices);
    inputVector timeStart(iTimeStart);
    inputVector timeEnd(iTimeEnd);
    inputVector variances(iVariances);
    inputVector varCenters(iVarCenters);

    MyMatrix exoData;
    if (hasExo)
      {
	if (!Rf_isNull(iExoData))
	  {
	    exoData = Rcpp::as<MyMatrix>(iExoData);
	    assert(exoData.n_cols == uint(exo.nrow()));
	  }
	else
	  hasExo = false;
      }
    
    MyMatrix paramMat = RtoMatrix(param);

    try{
      OUsimulation ou(paramMat,
		      timeIndices,
		      processIndexSize,
		      houseIndexSize,
		      hasStart,
		      hasExo,
		      varByHouse,
		      varCenters,
		      start,
		      mean,
		      extraMean,
		      reversion,
		      simultaneous,
		      sigma,
		      extraVarAdd,
		      extraVarMult,
		      extraVarPow,
		      exo
		      );

      PutRNGstate();

      if (timeStart.size() > 0)
	return Rcpp::List::create(Rcpp::Named("instantaneous.sample") = ou.getSampleCurrent(),
				  Rcpp::Named("noninstantaneous.sample") = ou.getSampleHist(),
				  Rcpp::Named("obs") = 
				  ou.getSampleData(processIndices,
						   seriesIndices,
						   houseIndices,
						   timeStart,
						   timeEnd,
						   exoData,
						   variances),
				  Rcpp::Named("v") = variances,
				  Rcpp::Named("t1") = timeStart,
				  Rcpp::Named("t2") = timeEnd);
      else
	return Rcpp::List::create(Rcpp::Named("instantaneous.sample") = ou.getSampleCurrent(),
				  Rcpp::Named("noninstantaneous.sample") = ou.getSampleHist());
    }
    catch (const std::runtime_error& error)
      {
	PutRNGstate();
	return Rcpp::List::create();
      }
}
